import json
import os
import plotly.graph_objs as go
import plotly.io as pio
import numpy as np
import argparse

from ftio.plot.units import set_unit
from ftio.plot.helper import format_plot_and_ticks

def load_json_and_plot(filenames):
    # Get the current directory
    current_directory = os.getcwd()
    
    # Create a list to hold traces
    traces = []
    unit = "MB"
    order = 1e-6
        
    for filename in filenames:
        # Define the path to the JSON file
        json_file_path = os.path.join(current_directory, filename)
        
        # Load the JSON file
        with open(json_file_path, "r") as json_file:
            data = json.load(json_file)
        
        # Extract arrays from the JSON data
        b = np.array(data.get("b", []))
        t = np.array(data.get("t", []))

        
        if filenames.index(filename) == 0:
            unit, order = set_unit(b)
        
        # Create a scatter plot trace of b against t
        trace = go.Scatter(
            x=t,
            y=b * order,
            mode='lines+markers',
            line={"shape": "hv"},
            fill="tozeroy",
            name=filename  # Set the legend name to the filename
        )
        
        # Add the trace to the list
        traces.append(trace)
    
    # Create a figure and add all traces
    fig = go.Figure(data=traces)
    
    # Update layout for better visualization
    fig.update_layout(
        title="Stage-out Data",
        xaxis_title="Time (s)",
        yaxis_title=f"Bandwidth ({unit})",
        showlegend=True
    )
    fig = format_plot_and_ticks(fig)
    
    # Show the plot
    pio.show(fig)

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Load JSON data from multiple files and plot bandwidth vs. time.")
    parser.add_argument(
        'filenames',
        type=str,
        nargs='+',
        default=["bandwidth.json"],
        help="The paths to the JSON files to plot. Multiple files can be provided."
    )
    args = parser.parse_args()
    
    load_json_and_plot(args.filenames)



import sys
from ftio.api.metric_proxy.helper import extract_data
from ftio.api.metric_proxy.parse_proxy import load_proxy_trace_stdin, parse_all
from ftio.prediction.helper import print_data
from ftio.api.metric_proxy.helper import data_to_json
from ftio.prediction.tasks import ftio_task_save





def main():
    argv = sys.argv[1:]
    metrics = load_proxy_trace_stdin(deriv_and_not_deriv=False)
    ranks = 32

    # Make sure we do no plots
    argv.extend(["-e", "no"])

    data = []

    for metric, arrays in metrics.items():
        ftio_task_save(data, metric, arrays, argv, ranks, False)
    data_to_json(data)


if __name__ == "__main__":
    main()


""" Function for DTW calcluations
"""
from scipy.spatial.distance import euclidean
from fastdtw import fastdtw
import numpy as np
from threading import Thread



##Fill DTW Matrix
def fill_dtw_cost_matrix(s1, s2):
    l_s_1, l_s_2 = len(s1), len(s2)
    cost_matrix = np.zeros((l_s_1 + 1, l_s_2 + 1))
    for i in range(l_s_1 + 1):
        for j in range(l_s_2 + 1):
            cost_matrix[i, j] = np.inf
    cost_matrix[0, 0] = 0

    for i in range(1, l_s_1 + 1):
        for j in range(1, l_s_2 + 1):
            cost = abs(s1[i - 1] - s2[j - 1])
            # take last min from the window
            prev_min = np.min(
                [
                    cost_matrix[i - 1, j],
                    cost_matrix[i, j - 1],
                    cost_matrix[i - 1, j - 1],
                ]
            )
            cost_matrix[i, j] = cost + prev_min

    return cost_matrix[-1, -1]


##Call DTW function


def fdtw(s1, s2):
    distance, path = fastdtw(s1, s2, dist=euclidean)
    return distance, path



def evaluate_dtw(discret_arr, original_discret_signal, freq):
    dtw_k1, _ = fastdtw(discret_arr, original_discret_signal, dist=euclidean)
    print("    '-> \033[1;32mfreq %.2f Hz\033[1;0m --> dtw: %d" % (freq, dtw_k1))



def threaded_dtw(
    sum_all_components, df, dominant_X1, dominant_k1, dominant_X2, dominant_k2
):
    threads = []
    print("    '-> \033[1;35mCalculating DTW\033[1;0m")
    threads.append(
        Thread(
            target=evaluate_dtw,
            args=(
                dominant_X1,
                sum_all_components,
                df.iloc[dominant_k1]["freq"],
            ),
        )
    )
    threads.append(
        Thread(
            target=evaluate_dtw,
            args=(
                dominant_X2,
                sum_all_components,
                df.iloc[dominant_k2]["freq"],
            ),
        )
    )
    # t.append(Thread(target=evaluate_dtw, args=(dominant_X3,sum,df.iloc[dominant_k3]['freq'])))
    for thread in threads:
        thread.start()

    return threads



from datetime import datetime
import multiprocessing
import subprocess
import time
import os
import re
from rich.console import Console
from rich.panel import Panel
from rich.status import Status
from rich.markup import escape
from ftio.api.gekkoFs.jit.jitsettings import JitSettings
from ftio.api.gekkoFs.jit.setup_helper import (
    check,
    flaged_mpiexec_call,
    jit_print,
    get_pid,
)

console = Console()


def execute_block(call: str, raise_exception: bool = True, dry_run=False) -> str:
    """Executes a call and blocks till it is finished

    Args:
        call (str): bash call to execute
    """
    jit_print(f">> Executing {call}")
    if dry_run:
        print(call)
        return ""

    out = ""
    try:
        # process = subprocess.Popen(call, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        # remove capture_output=True to unblock
        out = subprocess.run(
            call,
            shell=True,
            text=True,
            capture_output=True,
            check=True,
            executable="/bin/bash",
        )
        out = out.stdout
    except subprocess.CalledProcessError as e:
        error_message = (
            f"[red]Command failed:[/red] {call}\n"
            f"[red]Exit code:[/red] {e.returncode}\n"
            f"[red]Output:[/red] {e.stdout.strip()}\n"
            f"[red]Error:[/red] {e.stderr.strip()}"
        )
        console.print(f"[red]{error_message}\n[/]")
        if raise_exception:
            raise

    return out


def execute_block_and_log(call: str, log_file: str) -> float:
    """Executes a call and logs it's output. This is a blocking call that
    writes the output to the log once finished.

    Args:
        call (str): bash call to execute
        log_file (str): absolute location of the log file

    Returns:
        float: execution time of the call
    """
    log_message = f">> Executing command: {call}\n"
    jit_print("[cyan]" + log_message)
    start = time.time()
    end = start
    try:
        out = subprocess.run(
            call,
            shell=True,
            capture_output=True,
            text=True,
            check=True,
            executable="/bin/bash",
        )
        end = time.time()
        log_message += f"Output:\n{out.stdout}\n"
    except subprocess.CalledProcessError as e:
        log_message += f"Error:\n{e.stderr}\n"
        error_message = (
            f"[red]Command failed:[/red] {call}\n"
            f"[red]Exit code:[/red] {e.returncode}\n"
            f"[red]Output:[/red] {e.stdout.strip()}\n"
            f"[red]Error:[/red] {e.stderr.strip()}"
        )
        console.print(f"[red]{error_message}[/]")
        raise
    finally:
        # Write the log message to the file
        with open(log_file, "a") as file:
            file.write(log_message)
    return end - start


def execute_block_and_monitor(
    verbose: bool, call: str, log_file: str = "", log_err_file: str = "", dry_run=False
):
    if len(log_err_file) == 0:
        log_err_file = log_file

    process = execute_background(call, log_file, log_err_file, dry_run)
    if verbose:
        out = monitor_log_file(log_file, "")
        if log_err_file != log_file:
            err = monitor_log_file(log_err_file, "")

    _ = process.communicate()
    if verbose:
        out.terminate()
        if log_err_file != log_file:
            err.terminate()


def execute_background(
    call: str, log_file: str = "", log_err_file: str = "", dry_run=False
):
    """executes a call in the background and sets up a log dir

    Args:
        call (str): call to execute
        log_dir (str, optional): log dir directory. Defaults to "".
        error_dir (str, optional): error die directory. Defaults to "".

    Returns:
        _type_: _description_
    """
    jit_print(f"[cyan]>> Executing {call}")
    with open(log_file, "a") as file:
        file.write(f">> Executing {call}")

    if dry_run:
        print(call)
        call = ""
        return subprocess.Popen(call, shell=True, executable="/bin/bash")

    # if log_file and log_err_file:
    #     call = f"{call} >> {log_file} 2>> {log_err_file}"
    # elif log_file:
    #     call = f"{call} >> {log_file}"
    # else:
    #     pass
    # print(call)
    # process = subprocess.Popen(call, shell=True, preexec_fn=os.setpgrp,stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    if log_file and log_err_file:
        with open(log_file, "a") as log_out:
            with open(log_err_file, "w") as log_err:
                process = subprocess.Popen(
                    call,
                    shell=True,
                    executable="/bin/bash",
                    stdout=log_out,
                    stderr=log_err,
                )
    elif log_file:
        with open(log_file, "a") as log_out:
            process = subprocess.Popen(
                call, shell=True, executable="/bin/bash", stdout=log_out, stderr=log_out
            )
    else:
        process = subprocess.Popen(
            call,
            shell=True,
            executable="/bin/bash",
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
    return process


def execute_background_and_log(
    settings: JitSettings, call: str, log_file: str, name="", err_file: str = ""
) -> subprocess.Popen:
    """execute call in background and returns process. The output is displayed using a
    thread that reads the log file

    Args:
        settings (JitSettings): jit settings
        call (str): bash call to execute
        log_file (str): absolute location of the log file to monitor
        name (str, optional): The src of the file. If set to daemon, proxy, or ftio, the output
        is colored. Defaults to "".. Defaults to "".

    Returns:
        subprocess.Popen: process
    """
    process = execute_background(call, log_file, err_file, settings.dry_run)
    get_pid(settings, name, process.pid)

    _ = monitor_log_file(log_file, name)
    return process


def execute_background_and_log_in_process(
    call: str, log_file: str, name="", err_file: str = "", dry_run=False
):
    """execute call in background and returns process. The output is displayed using a
    thread that reads the log file

    Args:
        settings (JitSettings): jit settings
        call (str): bash call to execute
        log_file (str): absolute location of the log file to monitor
        name (str, optional): The src of the file. If set to daemon, proxy, or ftio, the output
        is colored. Defaults to "".. Defaults to "".

    Returns:
        subprocess.Popen: process
    """
    if dry_run:
        print(call)
        call = ""

    process = execute_background(call, log_file, err_file, dry_run)
    # get_pid(settings, name, process.pid)
    # daemon is noisy
    _ = monitor_log_file(log_file, src=name)
    _, stderr = process.communicate()
    if process.returncode != 0:
        console.print(f"[red]Error executing command:{call}")
        console.print(f"[red] Error was:\n{stderr}")
    else:
        pass


def execute_block_and_wait_line(call: str, filename: str, target_line: str) -> None:
    """executes a call and wait for a line to appear

    Args:
        call (str): bash call to execute
        filename (str): file to monitor for target_line to appear
        target_line (str): target line that need to appear in filename to
        stop the execution of this function
    """
    execute_block(call)
    _ = wait_for_line(filename, target_line)


def execute_background_and_wait_line(
    call: str, filename: str, target_line: str, dry_run: bool = False
) -> None:
    """executes a call and wait for a line to appear

    Args:
        call (str): bash call to execute
        filename (str): file to monitor for target_line to appear
        target_line (str): target line that need to appear in filename to
        stop the execution of this function
    """
    process = execute_background(call, filename, dry_run=dry_run)
    _ = monitor_log_file(filename, "")
    if not dry_run:
        _ = wait_for_line(filename, target_line)
        stdout, stderr = process.communicate()
        if process.returncode != 0:
            console.print(f"[red]Error executing command:{call}")
            console.print(f"[red] Error was:\n{stderr}")
        else:
            # console.print(stdout, style="bold green")
            pass


def monitor_log_file(file: str, src: str = "") -> multiprocessing.Process:
    """monitors a file and displays its output on the console. A process is
    in charge of monitoring the file.

    Args:
        file (str): absolute File path
        src (str, optional): The src of the file. If set to daemon, proxy, or ftio, the output
        is colored. Defaults to "".
    """
    monitor_process = multiprocessing.Process(target=print_file, args=(file, src))
    monitor_process.daemon = True
    monitor_process.start()

    return monitor_process


def end_of_transfer(
    settings: JitSettings, log_file: str, call: str, monitored_files: list[str] = []
) -> None:
    if settings.dry_run:
        return

    online = False
    stuck = False
    hits = 0
    limit = 100
    if len(monitored_files) == 0:
        monitored_files = get_files(settings, True)
        online = True
        stuck = True

    last_lines = read_last_n_lines(log_file)
    n = len(monitored_files)
    if "Transfer finished for []" in last_lines:
        # Nothing to move
        return
    elif n == 0:
        return
    else:
        with open(log_file, "r") as file:
            # Move to the end of the file
            file.seek(0, 2)
            last_pos = file.tell()
            buffer = ""
            with Status(
                f"[cyan]Waiting for end of transfer...monitoring {n} files",
                console=console,
            ) as status:
                while len(monitored_files) > 0:
                    time.sleep(0.1)  # Short sleep interval to quickly catch new lines

                    # Check if the file has grown
                    current_pos = file.tell()
                    if current_pos < last_pos:
                        # Log file was truncated or rotated, reset
                        file.seek(0, 2)
                        last_pos = file.tell()

                    # Read new content
                    file.seek(last_pos)
                    new_data = file.read()
                    last_pos = file.tell()

                    # Process the new data
                    if new_data:
                        buffer += new_data
                        lines = buffer.splitlines()

                        # Check if the last line is incomplete
                        if not new_data.endswith("\n"):
                            buffer = lines.pop()  # Save incomplete line in the buffer
                        else:
                            buffer = ""

                        check(settings)
                        # Process each line
                        for line in lines:
                            content = line.strip()
                            # jit_print(">>[cyan]**{content}")
                            if "Deleting" in content:
                                # Check if any of the monitored files are mentioned in the line
                                for name in monitored_files:
                                    if name in content:
                                        monitored_files.remove(name)
                                        jit_print(
                                            f"[cyan]>> finished moving '{name}'. Remaining files ({len(monitored_files)})",
                                            True,
                                        )
                                        console.print(
                                            f"Waiting for {len(monitored_files)} more files to be deleted: {monitored_files}"
                                        )
                                        jit_print(
                                            f"[cyan]>> Files in mount dir {monitored_files}",
                                            True,
                                        )
                                        status.update(
                                            f"Waiting for {len(monitored_files)} more files to be deleted: {monitored_files}"
                                        )
                                        hits = 0
                    if online:
                        # avoids dead loop for disappearing files
                        monitored_files = get_files(settings, False)
                        hits += 1
                        status.update(
                            f"Waiting for {len(monitored_files)} more files to be deleted [yellow]({hits}/{limit})[/]: {monitored_files}"
                        )
                        if hits > 4:
                            if stuck:
                                jit_print("[cyan]>> Stuck? Triggering cargo again\n")
                                _ = execute_background(
                                    call, settings.cargo_log, settings.cargo_err
                                )
                                stuck = False
                        if hits > limit:
                            jit_print("[cyan]>> Stopping stage out\n")
                            return
                # All monitored files have been processed
                timestamp = get_time()
                status.update(
                    f"\n[bold green]JIT [cyan]>> finished moving all files at  [{timestamp}]"
                )
                jit_print(
                    f"\n[cyan]>> finished moving all files at [{timestamp}]", True
                )


def end_of_transfer_online(
    settings: JitSettings, log_file: str, call: str, timeout=180
) -> None:
    if settings.dry_run:
        return

    repeated_trigger = True
    copy = False  # Trigger cargo again
    monitored_files = get_files(settings, True)
    stuck_time = 5
    last_lines = read_last_n_lines(log_file)
    stuck_files = 0
    n = len(monitored_files)
    if "Transfer finished for []" in last_lines:
        # Nothing to move
        return
    elif n == 0:
        return
    else:
        with Status(
            f"[cyan]Waiting for end of transfer...monitoring {n} files",
            console=console,
        ) as status:
            start_time = time.time()
            last_cargo_time = start_time
            hit = 0
            while len(monitored_files) > 0:
                time.sleep(0.1)  # Short sleep interval to quickly catch new lines
                # jit_print(f"[cyan]>> Files in mount dir {monitored_files}",True)
                status.update(
                    f"Waiting for {len(monitored_files)} files: {monitored_files}"
                )

                passed_time = int(time.time() - start_time)
                time_since_last_cargo = int(time.time() - last_cargo_time)

                if passed_time >= timeout:
                    status.update("Timeout reached")
                    jit_print("[bold red]>> Timeout reached[/]")
                    break
                if repeated_trigger:
                    hit += 1
                    # jit_print(f"{time_since_last_cargo} >= {stuck_time} and {stuck_files} == {len(monitored_files)}")
                    if time_since_last_cargo >= stuck_time and stuck_files == len(
                        monitored_files
                    ):
                        jit_print(
                            f"[cyan]>> Stucked for {stuck_time} sec. Triggering cargo again\n"
                        )
                        _ = execute_background(
                            call, settings.cargo_log, settings.cargo_err
                        )
                        last_cargo_time = time.time()
                        stuck_time = stuck_time * 2
                        jit_print(f"[cyan]>> Stucked increased to {stuck_time}\n")
                        jit_print(
                            f">> Waiting for {len(monitored_files)} more files to be deleted: {monitored_files}"
                        )
                        hit = 0
                        if "Transfer finished for []" in last_lines:
                            break
                    if hit == 3:
                        stuck_files = len(monitored_files)
                        hit = 0
                monitored_files = get_files(settings, False)

            timestamp = get_time()
            status.update(
                f"\n[bold green]JIT [cyan]>> finished moving all files at  [{timestamp}]"
            )
            jit_print(f"\n[cyan]>> finished moving all files at [{timestamp}]", True)


def get_files(settings: JitSettings, verbose=True):
    monitored_files = []
    files = ""
    # TODO: fix find for gekko 
    command_ls = flaged_mpiexec_call(settings, f" ls -l {settings.gkfs_mntdir}")
    # command_ls = flaged_mpiexec_call(settings, f" find {settings.gkfs_mntdir}")
    try:
        # files = subprocess.check_output(command_ls, shell=True).decode()
        files = subprocess.check_output(command_ls, shell=True, text=True)
        if files:
            files = files.splitlines()
            files = [f for f in files if "LIBGKFS" not in f]
            files = [
                file.replace(f"{settings.gkfs_mntdir}", "")
                for file in files
                if file.replace(f"{settings.gkfs_mntdir}", "")
            ]
            # remove directories
            files = [item for item in files if "." in item]

        monitored_files = files_filtered(files, settings.regex_match, verbose)
        if verbose:
            timestamp = get_time()
            jit_print(f"[cyan]>> Files that need to be stage out: [{timestamp}]")
            for f in monitored_files:
                console.print(f"[cyan]{f}[/]")
    except Exception as e:
        jit_print(f"[red] >> Error listing files script:\n{e}")

    return monitored_files


def get_time():
    return datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]


def files_filtered(list_of_files: list[str], regex_pattern, verbose=True) -> list[str]:
    monitored_files = []
    if list_of_files:
        if regex_pattern:
            # check if it contains file name without path (less error prone)
            while "/" in regex_pattern:
                regex_pattern = regex_pattern.split("/", 1)[1]
            if verbose:
                jit_print(f"[cyan]>> Cleaned Regex pattern to: {regex_pattern} ")
        regex = re.compile(regex_pattern)

        for f in list_of_files:
            if regex.match(f):
                # jit_print(f"[yellow]>> Ignoring: {f} ")
                monitored_files.append(f)

    return monitored_files


def print_file(file, src=""):
    """Continuously monitor the log file for new lines and print them."""
    color = ""
    close = ""
    wait_time = 0.05
    if src:
        if "daemon" in src.lower():
            color = "[purple4]"
            wait_time = 0.1
        elif "proxy" in src.lower():
            color = "[deep_pink1]"
            wait_time = 0.1
        elif any(keyword in src.lower() for keyword in ["dlio","lammp"]):
            color = "[gold3]"
            wait_time = 0.1
        elif "ftio" in src.lower():
            color = "[deep_sky_blue1]"
            wait_time = 0.1
        elif "error" in src.lower():
            color = "[red]"
            wait_time = 0.1

        if color:
            close = "[/]"

    while not os.path.exists(file):
        if "error" in src.lower():
            time.sleep(0.1)
        else:
            with console.status(
                f"[bold green]Waiting for {file} to appear ..."
            ) as status:
                time.sleep(0.1)

    with open(file, "r") as file:
        # Go to the end of the file
        file.seek(0, os.SEEK_END)
        buffer = []
        last_print_time = time.time()

        while True:
            line = file.readline()
            if line:
                buffer.append(line.rstrip())
            else:
                # If there's no new line, wait briefly
                time.sleep(wait_time)

            # Group and print the buffered lines every 0.1 seconds
            current_time = time.time()
            if current_time - last_print_time >= wait_time and buffer:
                # Print grouped lines
                content = "\n".join(buffer)
                buffer.clear()
                last_print_time = current_time

                if not src or "cargo" in src:
                    print(content)
                else:
                    console.print(
                        Panel.fit(
                            color + escape(content) + close,
                            title=src.capitalize(),
                            style="white",
                            border_style="white",
                            title_align="left",
                        )
                    )


def wait_for_file(filename: str, timeout: int = 180, dry_run=False) -> None:
    """Waits for a file to be created

    Args:
        file (str): absolute file path
    """
    if dry_run:
        return

    start_time = time.time()
    with Status(
        f"[cyan]Waiting for {filename} to be created...\n", console=console
    ) as status:
        while not os.path.isfile(filename):
            passed_time = int(time.time() - start_time)
            if passed_time >= timeout:
                status.update("Timeout reached")
                jit_print("[bold red]>> Timeout reached[/]")
                return

            status.update(
                f"[cyan]Waiting for {filename} to be created... ({passed_time}/{timeout}) s"
            )
            time.sleep(0.1)  # Wait for 1 second before checking again

        # When the file is created, update the status
        status.update(f"{filename} has been created.")
        jit_print(f">> {filename} has been created.", True)


def wait_for_line(
    filename: str, target_line: str, msg: str = "", timeout: int = 60, dry_run=False
) -> bool:
    """
    Waits for a specific line to appear in a log file

    Args:
        filename (str): The path to the log file.
        target_line (str): The line of text to wait for.
        msg (str): Message to print
        timeout (int): maximal timeout
    """
    success = True
    if dry_run:
        return success
    start_time = time.time()
    if not msg:
        msg = "Waiting for line to appear..."

    while not os.path.exists(filename):
        with console.status(
            f"[bold green]Waiting for {filename} to appear ..."
        ) as status:
            time.sleep(0.1)

    with open(filename, "r") as file:
        # Move to the end of the file to start monitoring
        # file.seek(0, 2)  # Go to the end of the file and look at the last 10 entris
        try:
            file.seek(-2, os.SEEK_END)
        except:
            file.seek(0, 0)  # Go to the end of the file and look at the last 10 entris

        with Status(f"[cyan]{msg}", console=console) as status:
            while True:
                line = file.readline()
                if not line:
                    # If no line, wait and check againet
                    time.sleep(0.01)
                    passed_time = int(time.time() - start_time)
                    if passed_time >= timeout:
                        status.update("Timeout reached.")
                        jit_print("[bold red]>> Timeout reached[/]", True)
                        success = False
                        break
                    status.update(f"[cyan]{msg} ({passed_time}/{timeout})")
                    continue

                if target_line in line:
                    timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]
                    status.update(f"Found target line: '{target_line}'")
                    jit_print(f"[cyan]>> Stopped waiting at [{timestamp}]", True)
                    break
        return success


def read_last_n_lines(filename, n=3):
    """Read the last `n` lines of a file."""
    with open(filename, "rb") as f:
        # Move the pointer to the end of the file
        f.seek(0, 2)
        file_size = f.tell()

        # Initialize variables
        lines = []
        buffer = b""
        pointer = file_size

        while pointer > 0 and len(lines) < n:
            # Move the pointer backwards by one byte and read
            pointer -= 1
            f.seek(pointer)
            buffer = f.read(1) + buffer

            # If a newline is found, add the line to the list
            if buffer.startswith(b"\n") or pointer == 0:
                lines.append(buffer.decode("utf-8"))
                buffer = b""

        return lines[-n:]



"""Performs prediction with Pools (ProcessPoolExecutor) and a callback mechanism"""
from __future__ import annotations
import zmq
import subprocess
import time
from ftio.prediction.async_process import join_procs
from ftio.prediction.processes import prediction_process
from ftio.prediction.helper import print_data, export_extrap
from ftio.prediction.async_process import handle_in_process
from ftio.parse.args import parse_args
from ftio.freq.helper import MyConsole

CONSOLE = MyConsole()
CONSOLE.set(True)

def predictor_with_processes_zmq(
    data, queue, count, hits, start_time, aggregated_bytes, args, b_app, t_app
)-> None:
    """performs prediction in ProcessPoolExecuter. FTIO is a submitted future and probability is calculated as a callback

    Args:   
        filename (str): name of file
        data (Manager().list): List of dicts with all predictions so far
        queue (Manager().Queue): queue for FTIO data
        count (Manager().Value): number of prediction
        hits (Manager().Value): hits indicating how often a dominant frequency was found
        start_time (Manager().Value): start time window for ftio
        aggregated_bytes (Manager().Value): total bytes transferred so far
        args (list[str]): additional arguments passed to ftio
    """
    procs = []
    context = zmq.Context()
    socket = context.socket(socket_type=zmq.PULL)
    #parse arguments
    tmp_args = parse_args(args)
    addr = tmp_args.zmq_address
    port = tmp_args.zmq_port

    # bind the socket
    socket = bind_socket(addr,port)
    # can be extended to listen to multiple sockets
    poller = zmq.Poller()
    poller.register(socket, zmq.POLLIN)
    
    if '-zmq' not in args:
        args.extend(['--zmq'])
    
    # Loop and predict if changes occur
    try:
        with CONSOLE.status("[green] started\n",spinner="arrow3") as status:
            while True:
                # join procs
                procs = join_procs(procs)
                # get messages
                msgs, ranks = receive_messages(socket, poller)

                if not msgs:
                    CONSOLE.print("[red]No messages[/]")
                    continue
                CONSOLE.print(f"[cyan]Got message from {ranks}:[/]")
                status.update("")

                # launch prediction_process
                # TODO: append b_app and t_app like predictor_gekko_zmq use the flag --zmq to indicate this
                # put all in msgs and call it zmq_data
                procs.append(
                    handle_in_process(
                        prediction_process,
                        args=(data, queue, count, hits, start_time, aggregated_bytes, args, msgs)
                    )
                )
    except KeyboardInterrupt:
        print_data(data)
        export_extrap(data)
        print('-- done -- ')



def bind_socket(addr, port):
    """Bind the ZMQ socket, retrying with a corrected IP if necessary."""
    context = zmq.Context()
    socket = context.socket(zmq.PULL)
    try:
        socket.bind(f"tcp://{addr}:{port}")
    except zmq.error.ZMQError as e:
        CONSOLE.print(f"[yellow]Error encountered:\n{e}[/]")
        CONSOLE.print("[yellow]Wrong IP address. Attempting to correct...[/]")
        addr = str(subprocess.check_output("ip addr | grep 'inet 10' | awk  '{print $2}'", shell=True))
        end   = addr.rfind("/")
        start = addr.find("'")
        addr = addr[start+1:end]
        CONSOLE.print("[bold green]Corrected IP address:[/]", addr)
        socket.bind(f"tcp://{addr}:{port}")

    CONSOLE.print(f"[green]FTIO is running on: {addr}:{port}[/]")

    return socket



def receive_messages(socket, poller):
    """Polls for and receives messages from the socket, returning a list of messages and count."""
    msgs = []
    ranks = 0
    # start = time.time()
    socks = dict(poller.poll(1000))
    #1) just a single msg
    # msg = socket.recv(zmq.NOBLOCK)  

    #2) Loop and accept messages from both channels, acting accordingly
    # if socks:
    #     if socks.get(socket) == zmq.POLLIN:
    #         print(f"got message ",{socket.recv(zmq.NOBLOCK)})
    # else:
    #     print("No message received")
    #     continue
    
    #3) Loop and accept messages from both channels, acting accordingly
    while socks: 
        if socks.get(socket) == zmq.POLLIN:
            msgs.append(socket.recv(zmq.NOBLOCK))
            # CONSOLE.print(f"[cyan]Got message {ranks}:[/] {msg}")
            ranks += 1
        # if time.time() - start > 0.5:
        #     break
        socks = dict(poller.poll(1000))
    
    return msgs, ranks


import json
import pandas as pd 
import numpy as np
from rich.progress import (
    Progress,
    SpinnerColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
    TaskProgressColumn,
    TextColumn,
    BarColumn,
)

def extract_data(data):
    """Extracts relevant data that is not NaN 

    Args:
        data (_type_): _description_

    Returns:
        _type_: _description_
    """
    # Prepare the data for the plot
    data_points = []

    for d in data:
        if len(d['dominant_freq']) > 0 and len(d['conf']) > 0:
            max_conf_index = np.argmax(d['conf'])
            dominant_freq = d['dominant_freq'][max_conf_index]
            conf = d['conf'][max_conf_index]*100
            phi = d['phi'][max_conf_index] #np.degrees(d['phi'][max_conf_index])
            amp = d['amp'][max_conf_index]
            t_s = d['t_start']
            t_e = d['t_end']
            data_points.append((d['metric'], dominant_freq, conf, amp, phi, t_s, t_e))
        else:
            continue 
            data_points.append((d['metric'], np.NaN, np.NaN, np. NaN, np.NaN, np. NaN))

    # Create a DataFrame for the plot
    df = pd.DataFrame(data_points, columns=['Metric', 'Dominant Frequency', 'Confidence', 'Amp', 'Phi', 'time start', 'time end'])
    df.sort_values(by='Dominant Frequency',inplace=True)

    return df



class NpArrayEncode(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            np.nan_to_num(x=obj, copy=False, nan=0.0, posinf=0.0, neginf=0.0)
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


def data_to_json(data: list[dict]) -> None:
    print(json.dumps(data, cls=NpArrayEncode))


def create_process_bar(total_files):

    
    # Create a progress bar
    progress = Progress(
        SpinnerColumn(),
        TextColumn(
            "[progress.description]{task.description} ({task.completed}/{task.total})"
        ),
        BarColumn(),
        TaskProgressColumn(),
        TimeRemainingColumn(),
        "[yellow]-- runtime",
        TimeElapsedColumn(),
    )
    return progress


import sys
import subprocess
import getopt
import os
import signal
import shutil
from rich.console import Console
from ftio.api.gekkoFs.jit.jitsettings import JitSettings
from ftio.api.gekkoFs.jit.jittime import JitTime

console = Console()


def is_port_in_use(port_number):
    try:
        # Run the netstat command and search for the port number
        netstat_output = subprocess.check_output(
            f"netstat -tlpn | grep ':{port_number} '",
            shell=True,
            stderr=subprocess.STDOUT,
        ).decode("utf-8")
        # If output is found, the port is in use
        if netstat_output:
            jit_print(
                f"[bold red]>> Error: Port {port_number} is already in use...[/bold red]"
            )
            return True  # Port is in use
    except subprocess.CalledProcessError:
        # grep returns a non-zero exit status if it finds no matches
        jit_print(f"[bold blue]>> Port {port_number} is available.[/bold blue]")
        return False  # Port is free


def check_port(settings: JitSettings):
    """Check if a port is available and terminate any existing process using it."""
    if is_port_in_use(settings.port):
        jit_print(
            f"[bold red]>> Error: Port {settings.port} is already in use on {settings.address_ftio}. Terminating existing process...[/]"
        )

        # Identify the process ID using the settings.port
        try:
            process_output = (
                subprocess.check_output(
                    f"netstat -nlp | grep :{settings.port} ", shell=True
                )
                .decode()
                .strip()
            )
            process_id = process_output.split()[6].split("/")[0]

            if process_id:
                jit_print(
                    f"[bold yellow]>> Terminating process with PID: {process_id}[/]"
                )
                os.kill(int(process_id), 9)
                jit_print(
                    f"[bold green]>> Using port {settings.port} on {settings.address_ftio}.[/]"
                )
                return 0
            else:
                jit_print(
                    f">>[bold red]>> Failed to identify process ID for PORT {settings.port}.[/]"
                )
                exit(1)
        except subprocess.CalledProcessError as e:
            jit_print(f"[bold red]>> Failed to retrieve process information: {e}[/]")
            exit(1)
    else:
        jit_print(
            f"[bold green]>> Using port {settings.port} on {settings.address_ftio}.[/]"
        )


def parse_options(settings: JitSettings, args: list) -> None:

    try:
        opts, args = getopt.getopt(
            args,
            "a:r:n:p:c:o:t:j:l:i:e:xdvyuh",
            [
                "address=",
                "port=",
                "nodes=",
                "total_procs=",
                "procs_list=",
                "omp_threads=",
                "max-time=",
                "job-id=",
                "log-name=",
                "install_location=",
                "exclude=",
                "exclude-all",
                "dry_run",
                "verbose",
                "skip_confirm",
                "use-mpirun",
                "help",
            ],
        )
    except getopt.GetoptError as err:
        console.print(f"[bold red]Error: {err}[/bold red]")
        sys.exit(2)

    for opt, arg in opts:
        if opt in ("-a", "--address"):
            settings.address_ftio = arg
        elif opt in ("-r", "--port"):
            settings.port = arg
        elif opt in ("-n", "--nodes"):
            settings.nodes = int(arg)
        elif opt in ("-t", "--max-time"):
            settings.max_time = int(arg)
        elif opt in ("-j", "--job-id"):
            settings.job_id = int(arg)
            settings.static_allocation = True
        elif opt in ("-l", "--log-name"):
            settings.log_dir = arg
        elif opt in ("-i", "--install_location"):
            settings.install_location = arg
            install_all(settings)
        elif opt in ("-c", "--total_procs"):
            settings.procs = int(arg)
        elif opt in ("-o", "--omp_threads"):
            settings.omp_threads = int(arg)
        elif opt in ("-p", "--procs_list"):
            # Split the argument by comma to get the list of numbers
            procs_list = arg.split(",")
            try:
                procs_list = [int(proc) for proc in procs_list]
            except ValueError:
                console.print(
                    "[bold red]Invalid --procs value. It must be a comma-separated list of numbers.[/]"
                )
                sys.exit(1)

            # Check the number of elements and assign accordingly
            if len(procs_list) > 5:
                console.print("[bold red]Too many values for --procs. Maximum is 5.[/]")
                sys.exit(1)
            elif len(procs_list) > 0:
                settings.procs_app = int(procs_list[0])
            if len(procs_list) > 1:
                settings.procs_daemon = int(procs_list[1])
            if len(procs_list) > 2:
                settings.procs_proxy = int(procs_list[2])
            if len(procs_list) > 3:
                settings.procs_cargo = int(procs_list[3])
            if len(procs_list) > 4:
                settings.procs_ftio = int(procs_list[4])

        elif opt in ("-e", "--exclude"):
            jit_print("[bold yellow]>> Excluding: [/]")
            if not arg or arg.startswith("-"):
                settings.exclude_ftio = True
                console.print("[yellow]- ftio[/]")
            else:
                excludes = arg.split(",")
                for exclude in excludes:
                    if exclude.lower() == "ftio":
                        settings.exclude_ftio = True
                        console.print("[yellow]- ftio[/]")
                    elif exclude.lower() == "cargo":
                        settings.exclude_cargo = True
                        console.print("[yellow]- cargo[/]")
                    elif exclude.lower() in ("gkfs", "daemon", "proxy"):
                        if exclude.lower() == "gkfs":
                            settings.exclude_daemon = True
                            settings.exclude_proxy = True
                            console.print("[yellow]- gkfs[/]")
                        elif exclude.lower() == "daemon":
                            settings.exclude_daemon = True
                            console.print("[yellow]- daemon[/]")
                        elif exclude.lower() == "proxy":
                            settings.exclude_proxy = True
                            console.print("[yellow]- proxy[/]")
                    elif exclude.lower() == "all":
                        settings.exclude_all = True
                        console.print("[yellow]- all[/]")
                    else:
                        jit_print(f"[bold red]>> Invalid exclude option: {exclude} [/]")
                        sys.exit(1)
        elif opt in ("-x", "--exclude-all"):
            settings.exclude_all = True
        elif opt in ("-d", "--dry_run"):
            settings.dry_run = True
        elif opt in ("-v", "--verbose"):
            settings.verbose = True
        elif opt in ("-y", "--skip_confirm"):
            settings.skip_confirm = True
        elif opt in ("-u", "--use-mpirun"):
            settings.use_mpirun = True
        elif opt in ("-h", "--help"):
            error_usage(settings)
            sys.exit(1)

        else:
            jit_print(f"[bold red]>>Invalid option: {opt}[/]")
            error_usage(settings)
            sys.exit(1)

    settings.update()


def error_usage(settings: JitSettings):
    console.print(
        f"""
[bold]Usage: {sys.argv[0]} [OPTION] ... [/]

    -a | --address: X.X.X.X <string>
        default: [bold yellow]{settings.address_ftio}[/]
        Address where FTIO is executed. On a cluster, this is found 
        automatically by determining the address of node where FTIO 
        runs.

    -r | --port: XXXX <int>
        default: [bold yellow]{settings.port}[/]
        port for FTIO and GekkoFS.

    -n | --nodes: X <int>
        default: [bold yellow]{settings.nodes}[/]
        number of nodes to run the setup. In cluster mode, FTIO is 
        executed on a single node, while the rest (including the
        application) get X-1 nodes.

    -c | --total_procs: X <int>
        default: [bold yellow]{settings.procs}[/]
        if procs_list is skipped, this is the default number of procs assigned to all
        
    -p | --procs_list: x,x,..,x <list>
        default: [bold yellow]{settings.procs_app},{settings.procs_daemon},{settings.procs_proxy},{settings.procs_cargo},{settings.procs_ftio}[/]
        List of task per node/cpu per proc for app, daemon, proxy, cargo, and ftio, respectively.
        Assignment is from right to left depending on the length of the list.
        FTIO, GekkoFS (proxy and daemon) always have 1 task per node. The 
        assignment in this list specifics the cpu per task. For cargo and the app, the task per node
        is calculated as nodes*procs_cargo or nodes*procs_app, respectively.

    -o | --omp_threads: X <int>
        default: [bold yellow]{settings.omp_threads}[/]
        OpenMP threads used

    -t | --max-time: X <int>
        default: [bold yellow]{settings.max_time}[/]
        max time for the execution of the setup in minutes.

    -j | --job-id: X <int>
        default: [bold yellow] Auto detected[/]
        Skips allocating new resources and uses job id.
    
    -l | --log-name: <str>
        default: Auto set to number of nodes and job ID
        if provided, sets the name of the directory where the logs are stored.

    -e | --exclude: <str>,<str>,...,<str>
        default: ftio
        If this flag is provided, the setup is executed without the tool(s).
        Supported options include: ftio, daemon, proxy, gkfs (daemon + proxy), 
        cargo, and all (same as -x).

    -x | --exclude-all
        default: [bold yellow]{settings.exclude_all}[/]
        If this flag is provided, the setup is executed without FTIO, 
        GekkoFs, and Cargo.


    -d | --dry_run 
        default: [bold yellow]{settings.dry_run}[/]
        If provided, the tools and the app are not executed

    -v | --verbose
        default: [bold yellow]{settings.verbose}[/]
        If provided, the tools output of each step is shown

    -y | --skip_confirm 
        default: [bold yellow]{settings.skip_confirm}[/]
        If this flag is provided, the setup automatically cancels running jobs 
        name JIT

    -u | --use-mpirun
        default: [bold yellow]{settings.use_mpirun}[/]
        If this flag is provided, the setup avoids using srun and 
        uses mpirun. Use -j JobID in combination with this flag, as 
        all calls are executed with the assumption you already 
        ssh to the host

    -i | --install_location: full_path <str>
        default: [bold yellow]{settings.install_location}[/]
        Installs everything in the provided directory.

---- exit ----
    """
    )


def abort():
    jit_print("[bold red] >>> Aborting installation[/]")
    exit(1)


def install_all(settings: JitSettings) -> None:
    with console.status("[bold green]Starting installation...") as status:
        try:
            # Create directory
            jit_print(">>> Creating directory")
            status.update("[bold green]JIT >>> Creating directory[/]", speed=30)
            os.makedirs(settings.install_location, exist_ok=True)

            # unload CUDA for Gekko
            command = 'module -t list 2>&1 | grep -q "CUDA" && module unload $(module -t list 2>&1 | grep "CUDA")'
            try:
                # Execute the bash command
                subprocess.run(command, shell=True, check=True)
                jit_print(">>> CUDA module unloaded successfully.")
            except subprocess.CalledProcessError:
                jit_print(">>> CUDA module not found or failed to unload.")

            # Clone GKFS
            jit_print(">>> Installing GKFS")
            if os.path.isdir(os.path.join(settings.install_location, "gekkofs")):
                jit_print(">>> GKFS exists, skipping installation")
            else:
                status.update("[bold green]JIT >>> Installing GKFS[/]", speed=30)
                subprocess.run(
                    [
                        "git",
                        "clone",
                        "--recurse-submodules",
                        "https://storage.bsc.es/gitlab/hpc/gekkofs.git",
                    ],
                    cwd=settings.install_location,
                    check=True,
                )
                os.chdir(os.path.join(settings.install_location, "gekkofs"))
                subprocess.run(["git", "pull", "--recurse-submodules"], check=True)
                os.chdir(settings.install_location)

                # Build GKFS
                jit_print(">>> Building GKFS")
                status.update("[bold green]JIT >>> Building GKFS[/]", speed=30)
                subprocess.run(
                    [
                        "gekkofs/scripts/gkfs_dep.sh",
                        "-p",
                        "default_zmq",
                        f"{settings.install_location}/iodeps/git",
                        f"{settings.install_location}/iodeps",
                    ],
                    check=True,
                )
                build_dir = os.path.join(settings.install_location, "gekkofs", "build")
                os.makedirs(build_dir, exist_ok=True)
                subprocess.run(
                    [
                        "cmake",
                        "-DCMAKE_BUILD_TYPE=Release",
                        f"-DCMAKE_PREFIX_PATH={settings.install_location}/iodeps",
                        "-DGKFS_BUILD_TESTS=OFF",
                        f"-DCMAKE_INSTALL_PREFIX={settings.install_location}/iodeps",
                        "-DGKFS_ENABLE_CLIENT_METRICS=ON",
                        "..",
                    ],
                    cwd=build_dir,
                    check=True,
                )
                subprocess.run(
                    ["make", "-j", "4", "install"], cwd=build_dir, check=True
                )

                jit_print(">>> GEKKO installed")
                status.update("[bold green]JIT >>> GEKKO installed[/]", speed=30)

            jit_print(">>> Installing Boost")
            status.update("[bold green]JIT >>> Getting Boost[/]", speed=30)
            if os.path.isfile(
                os.path.join(settings.install_location, "boost_1_84_0.tar.gz")
            ):
                print("File exists.")
                jit_print(">>> Boost tar exists, skipping download")
            else:
                try:
                    subprocess.run(
                        [
                            "wget",
                            "https://archives.boost.io/release/1.84.0/source/boost_1_84_0.tar.gz",
                        ],
                        cwd=settings.install_location,
                        check=True,
                    )
                except:
                    jit_print(">>> Wget failed, copying boost from shared")
                    subprocess.run(
                        [
                            "cp",
                            "/lustre/project/nhr-admire/shared/boost_1_84_0.tar.gz",
                            ".",
                        ],
                        cwd=settings.install_location,
                        check=True,
                    )
            if os.path.isdir(os.path.join(settings.install_location, "boost_1_84_0")):
                jit_print(">>> Boost exists, skipping installation")
            else:
                status.update(
                    "[bold green]JIT >>> Extracting boost from tar archive [/]",
                    speed=30,
                )
                subprocess.run(
                    ["tar", "-xzf", "boost_1_84_0.tar.gz"],
                    cwd=settings.install_location,
                    check=True,
                )
                status.update(
                    "[bold green]JIT >>> Installing Boost: bootstrap[/]", speed=30
                )
                build_dir = os.path.join(settings.install_location, "boost_1_84_0")
                os.chdir(build_dir)
                subprocess.run(f"{build_dir}/bootstrap.sh", cwd=build_dir, check=True)
                status.update(
                    "[bold green]JIT >>> Installing Boost: b2 install[/]", speed=30
                )
                subprocess.run(
                    [
                        f"{build_dir}/b2",
                        "install",
                        f"--prefix={settings.install_location}/iodeps",
                        "toolset=gcc",
                    ],
                    cwd=build_dir,
                    check=True,
                )
                status.update(
                    "[bold green]JIT >>> Installing Boost: b2 headers[/]", speed=30
                )
                subprocess.run(
                    [
                        f"{build_dir}/b2",
                        "headers",
                        f"--prefix={settings.install_location}/iodeps",
                    ],
                    cwd=build_dir,
                    check=True,
                )
                jit_print(">>> Boost installed")
                os.chdir(settings.install_location)
                status.update("[bold green]JIT >>> Boost installed[/]", speed=30)

            jit_print(">>> Installing Cereal")
            status.update("[bold green]JIT >>> Installing Cereal[/]", speed=30)
            subprocess.run(
                ["git", "clone", "https://github.com/USCiLab/cereal"],
                cwd=settings.install_location,
                check=True,
            )
            os.chdir(os.path.join(settings.install_location, "cereal"))
            build_dir = os.path.join(settings.install_location, "cereal", "build")
            os.makedirs(build_dir, exist_ok=True)
            subprocess.run(
                [
                    "cmake",
                    f"-DCMAKE_PREFIX_PATH={settings.install_location}/iodeps",
                    f"-DCMAKE_INSTALL_PREFIX={settings.install_location}/iodeps",
                    "..",
                ],
                cwd=build_dir,
                check=True,
            )
            subprocess.run(["make", "-j", "4", "install"], cwd=build_dir, check=True)
            jit_print(">>> Cereal installed")
            status.update("[bold green]JIT >>> Cereal installed[/]", speed=30)

            # Install Cargo Dependencies: Thallium
            jit_print(">>> Installing Thallium")
            status.update("[bold green]JIT >>> Installing Thallium[/]", speed=30)
            subprocess.run(
                ["git", "clone", "https://github.com/mochi-hpc/mochi-thallium"],
                cwd=settings.install_location,
                check=True,
            )
            os.chdir(os.path.join(settings.install_location, "mochi-thallium"))
            build_dir = os.path.join(
                settings.install_location, "mochi-thallium", "build"
            )
            os.makedirs(build_dir, exist_ok=True)
            subprocess.run(
                [
                    "cmake",
                    f"-DCMAKE_PREFIX_PATH={settings.install_location}/iodeps",
                    f"-DCMAKE_INSTALL_PREFIX={settings.install_location}/iodeps",
                    "..",
                ],
                cwd=build_dir,
                check=True,
            )
            subprocess.run(["make", "-j", "4", "install"], cwd=build_dir, check=True)
            jit_print(">>> Mochi-thallium installed")
            status.update("[bold green]JIT >>> Mochi-thallium installed[/]", speed=30)

            # Clone and Build Cargo
            jit_print(">>> Installing Cargo")
            status.update("[bold green]JIT >>> Installing Cargo[/]", speed=30)
            subprocess.run(
                ["git", "clone", "https://storage.bsc.es/gitlab/hpc/cargo.git"],
                cwd=settings.install_location,
                check=True,
            )
            os.chdir(os.path.join(settings.install_location, "cargo"))
            replace_line_in_file(
                "src/master.cpp", 332, f'  auto patternFile = "{settings.regex_file}";'
            )
            subprocess.run(
                ["git", "checkout", "marc/nek5000"],
                check=True,
            )
            build_dir = os.path.join(settings.install_location, "cargo", "build")
            os.makedirs(build_dir, exist_ok=True)
            subprocess.run(
                [
                    "cmake",
                    "-DCMAKE_BUILD_TYPE=Release",
                    f"-DCMAKE_PREFIX_PATH={settings.install_location}/iodeps",
                    f"-DCMAKE_INSTALL_PREFIX={settings.install_location}/iodeps",
                    "..",
                ],
                cwd=build_dir,
                check=True,
            )
            subprocess.run(["make", "-j", "4", "install"], cwd=build_dir, check=True)

            jit_print(">>> Cargo installed")
            status.update("[bold green]JIT >>> Cargo installed[/]", speed=30)

            # Build IOR
            jit_print(">>> Installing IOR")
            status.update("[bold green]JIT >>> Installing IOR[/]", speed=30)
            subprocess.run(
                ["git", "clone", "https://github.com/hpc/ior.git"],
                cwd=settings.install_location,
                check=True,
            )
            os.chdir(os.path.join(settings.install_location, "ior"))
            subprocess.run(["./bootstrap"], check=True)
            subprocess.run(["./configure"], check=True)
            subprocess.run(["make", "-j", "4"], check=True)

            jit_print(">> Installation finished")
            status.update("[bold green]JIT >> Installation finished[/]", speed=30)
            console.print("\n>> Ready to go <<")
            status.update("\n>> Ready to go <<", speed=30)
            console.print("Call: ./jit.sh -n NODES -t MAX_TIME")
            status.update("Call: ./jit.sh -n NODES -t MAX_TIME", speed=30)

        except subprocess.CalledProcessError as e:
            jit_print(f"[bold red] >>> Error encountered: {e}[/]")
            status.update(
                "[bold green]JIT [bold red] >>> Error encountered: {e}[/]", speed=30
            )
            abort()


def replace_line_in_file(file_path, line_number, new_line_content):
    try:
        # Read the existing file content
        with open(file_path, "r") as file:
            lines = file.readlines()

        # Replace the specific line (line_number - 1 because list indices start at 0)
        if line_number - 1 < len(lines):
            lines[line_number - 1] = new_line_content + "\n"
        else:
            raise IndexError(
                "Line number exceeds the total number of lines in the file."
            )

        # Write the modified content back to the file
        with open(file_path, "w") as file:
            file.writelines(lines)

        print(f"Successfully replaced line {line_number} in {file_path}.")

    except FileNotFoundError:
        print(f"Error: The file {file_path} does not exist.")
    except IndexError as e:
        print(f"Error: {e}")
    except Exception as e:
        print(f"An unexpected error occurred: {e}")


def cancel_jit_jobs(settings: JitSettings):
    if settings.job_id == 0:
        # Check if the hostname contains 'cpu' or 'mogon'
        hostname = subprocess.check_output("hostname", shell=True).decode().strip()
        if "cpu" in hostname or "mogon" in hostname:
            # Get the list of job IDs with the name "JIT"
            try:
                jit_jobs = (
                    subprocess.check_output(
                        "squeue --me --name=JIT --format=%A", shell=True
                    )
                    .decode()
                    .strip()
                    .split("\n")[1:]
                )
            except subprocess.CalledProcessError:
                jit_jobs = []

            if not jit_jobs:
                return

            # Print the list of JIT jobs
            job_list = "\n".join(jit_jobs)
            jit_print(
                "[bold yellow]>> The following jobs with the name 'JIT' were found:[/]"
            )
            console.print(job_list)

            if not settings.skip_confirm:
                # Prompt the user to confirm cancellation
                confirmation = (
                    input("Do you want to cancel all 'JIT' jobs? (yes/no): ")
                    .strip()
                    .lower()
                )
            else:
                confirmation = "yes"

            if confirmation in {"yes", "y", "ye"}:
                for job_id in jit_jobs:
                    subprocess.run(f"scancel {job_id}", shell=True, check=True)
                    jit_print(f"[bold cyan]>> Cancelled job ID {job_id}[/]")
                jit_print("[bold green]>> All 'JIT' jobs have been cancelled[/]")
            else:
                jit_print("[bold yellow]>> No jobs were cancelled[/]")


def relevant_files(settings: JitSettings):

    if settings.verbose:  # Mimicking checking for the number of arguments
        jit_print(f"[cyan]>> Setting up ignored files[/]")

    # Create or update the regex file with the settings.regex_match
    with open(settings.regex_file, "w") as file:
        file.write(f"{settings.regex_match}\n")

    if settings.verbose:
        jit_print(f"[cyan]>> Files that match {settings.regex_match} are ignored [/]")

    # Display the contents of the regex file
    with open(settings.regex_file, "r") as file:
        content = file.read()
    if settings.verbose:
        jit_print(f"[cyan]>> content of {settings.regex_file}: \n{content}[/]")


def reset_relevant_files(settings: JitSettings) -> None:
    if settings.cluster:
        jit_print(f"[cyan]>> Resetting ignored files[/]")
        # Write the default regex pattern to the file
        settings.regex_match = ".*"
        with open(settings.regex_file, "w") as file:

            file.write(f"{settings.regex_match}\n")
        # Optionally jit_print the contents of the regex file
        # with open(settings.regex_file, 'r') as file:
        #     content = file.read()
        # jit_print(f"[bold cyan]>> cat {settings.regex_file}: \n{content} [/]\n")


def total_time(log_dir):
    time_log_file = os.path.join(log_dir, "time.log")

    # Calculate total time from the log file
    try:
        with open(time_log_file, "r") as file:
            lines = file.readlines()
        total_time = sum(float(line.split()[1]) for line in lines if "seconds" in line)
    except FileNotFoundError:
        total_time = 0

    # Print and append total time to the log file
    result = f"\n JIT --> Total Time: {total_time}\n"
    console.print("[bold green]" + result + "[/]")
    with open(time_log_file, "a") as file:
        file.write(result)


def allocate(settings: JitSettings) -> None:
    settings.app_nodes = 1

    if settings.use_mpirun and settings.job_id == 0:
        jit_print(
            " [bold yellow]>> Warning, executing mpiexec on remote. We advice using using fixed allocation with mpirun\n"
        )

    if settings.cluster:
        # allocate if needed
        if settings.job_id == 0:
            # Allocating resources
            jit_print("[green] ####### Allocating resources[/]")

            call = f"salloc -N {settings.nodes} -t {settings.max_time} {settings.alloc_call_flags}"
            jit_print(f"[cyan] >> Executing: {call} [/]")
            # Execute the salloc command
            subprocess.run(call, shell=True, check=True)

            # Get JIT_ID
            try:
                result = subprocess.run(
                    f"squeue -o '%.18i %.9P %.8j %.8u %.2t %.10M %.6D %R' | grep '{settings.job_name}' | sort -k1,1rn | head -n 1 | awk '{{print $1}}'",
                    shell=True,
                    capture_output=True,
                    text=True,
                    check=True,
                )
                settings.job_id = int(result.stdout.strip())
            except subprocess.CalledProcessError:
                settings.job_id = 0
        else:
            console.print(
                f"[bold green]JIT [green] ####### Using allocation with id: {settings.job_id}[/]"
            )

        # Get NODES_ARR
        if settings.job_id:
            try:
                nodes_result = subprocess.run(
                    f"scontrol show hostname $(squeue -j {settings.job_id} -o '%N' | tail -n +2)",
                    shell=True,
                    capture_output=True,
                    text=True,
                    check=True,
                )
                nodes_arr = nodes_result.stdout.splitlines()
                # console.print(f"[bold green] ## Node res {nodes_result}[/]")
                # console.print(f"[bold green] ## Node res stdout{nodes_result.stdout}[/]")
                # console.print(f"[bold green] ## Node arr {nodes_arr}[/]")
                # console.print(f"[bold green] ## split{nodes_result.stdout.split()}[/]")
                if nodes_arr:
                    try:
                        nodes_arr = nodes_arr[: int(settings.nodes)]
                    except Exception as e:
                        print(e)
                        console.print(
                            f"[bold green]JIT [red] >> Unable to decrease number of nodes. Using {settings.nodes}[/]"
                        )
            except subprocess.CalledProcessError:
                nodes_arr = []

            # Write to hostfile_mpi
            with open(f"{settings.dir}/hostfile_mpi", "w") as file:
                file.write("\n".join(nodes_arr) + "\n")

            if nodes_arr:
                # Get FTIO node
                settings.ftio_node = nodes_arr[-1]
                # Get node for single commands
                settings.single_node = nodes_arr[0]

                if len(nodes_arr) > 1:
                    settings.ftio_node_command = f"--nodelist={settings.ftio_node}"
                    settings.app_nodes_command = f"--nodelist={','.join(n for n in nodes_arr if n != settings.ftio_node)}"
                    settings.single_node_command = f"--nodelist={settings.single_node}"
                    settings.app_nodes = len(nodes_arr) - 1
                    # print(nodes_arr)
                    # print(f"{','.join(n for n in nodes_arr if n != settings.ftio_node)}\n")

                    # Remove FTIO node from hostfile_mpi
                    with open(f"{settings.dir}/hostfile_mpi", "r") as file:
                        lines = file.readlines()
                    with open(f"{settings.dir}/hostfile_mpi", "w") as file:
                        file.writelines(
                            line for line in lines if line.strip() != settings.ftio_node
                        )

                jit_print(f"[green]>> JIT Job Id: {settings.job_id} [/]")
                jit_print(f"[green]>> Allocated Nodes: {len(nodes_arr)} [/]")
                jit_print(f"[green]>> FTIO Node: {settings.ftio_node} [/]")
                jit_print(
                    f"[green]>> APP Node command: {settings.app_nodes_command} [/]"
                )
                jit_print(
                    f"[green]>> FTIO Node command: {settings.ftio_node_command} [/]"
                )

                # Print contents of hostfile_mpi
                with open(f"{settings.dir}/hostfile_mpi", "r") as file:
                    hostfile_content = file.read()
                jit_print(
                    f"[cyan]>> content of {settings.dir}/hostfile_mpi: \n{hostfile_content} [/]"
                )
            settings.update_geko_files()
        else:
            jit_print("[bold red]>> JOB ID could not be retrieved[/]")


def get_pid(settings: JitSettings, name: str, pid: int):
    if settings.cluster:
        call = f"ps aux | grep 'srun' | grep '{settings.job_id}' | grep '{name}' | grep -v grep | tail -1 | awk '{{print $2}}'"
        res = subprocess.run(
            call, shell=True, check=True, capture_output=True, text=True
        )
        if res.stdout.strip():
            try:
                pid = int(res.stdout.strip())
            except:
                pid = res.stdout.strip()

    if name.lower() in "cargo":
        settings.cargo_pid = pid
        jit_print(f">> Cargo startup successful. PID is {pid}")
    elif name.lower() in "gkfs_daemon":
        settings.gekko_daemon_pid = pid
        jit_print(f">> Gekko daemon startup successful. PID is {pid}")
    elif name.lower() in "gkfs_proxy":
        settings.gekko_proxy_pid = pid
        jit_print(f">> Gekko proxy startup successful. PID is {pid}")
    elif name.lower() in "ftio" or name.lower() in "predictor_jit":
        settings.ftio_pid = pid
        jit_print(f">> FTIO startup successful. PID is {pid}")


def handle_sigint(settings: JitSettings):
    if settings.trap_exit:
        settings.trap_exit = False
        jit_print("[bold blue]>> Keyboard interrupt detected. Exiting script.[/]")
        info = (
            f"{settings.app_call} with {settings.nodes} nodes ({settings.log_suffix})"
        )
        jit_print(f"[bold blue]>> Killing Job: {info}.\n Exiting script.[/]")
        log_failed_jobs(settings, info)
        soft_kill(settings)
        hard_kill(settings)
        jit_print(">> Exciting\n")
        sys.exit(0)


def soft_kill(settings: JitSettings) -> None:
    if settings.soft_kill:
        jit_print("[bold green]####### Soft kill [/]", True)

        if not settings.exclude_ftio:
            try:
                shut_down(settings, "FTIO", settings.ftio_pid)
                jit_print("[bold cyan] >> killed FTIO [/]")
            except:
                jit_print("[bold cyan] >> Unable to soft kill FTIO [/]")

    if not settings.exclude_daemon:
        try:
            shut_down(settings, "GEKKO", settings.gekko_daemon_pid)
            jit_print("[bold cyan] >> killed GEKKO DEMON [/]")
        except:
            jit_print("[bold cyan] >> Unable to soft kill GEKKO DEMON [/]")

    if not settings.exclude_proxy:
        try:
            shut_down(settings, "GEKKO", settings.gekko_proxy_pid)
            jit_print("[bold cyan] >> killed GEKKO PROXY [/]")
        except:
            jit_print("[bold cyan] >> Unable to soft  kill GEKKO PROXY [/]")

    if not settings.exclude_cargo:
        try:
            shut_down(settings, "CARGO", settings.cargo_pid)
            jit_print("[bold cyan] >> killed CARGO [/]")
        except:
            jit_print("[bold cyan] >> Unable to soft kill CARGO [/]")

    jit_print(">> Soft kill finished")


def hard_kill(settings) -> None:
    if settings.hard_kill:
        jit_print(f"[bold green]####### Hard kill[/]", True)

        if settings.cluster and not settings.static_allocation:
            # Cluster environment: use `scancel` to cancel the job
            _ = subprocess.run(
                f"scancel {settings.job_id}",
                shell=True,
                text=True,
                capture_output=True,
                check=True,
                executable="/bin/bash",
            )
        else:
            # Non-cluster environment: use `kill` to terminate processes
            processes = [
                settings.gkfs_daemon,
                settings.gkfs_proxy,
                settings.cargo,
                f"{settings.ftio_bin_location}/predictor_jit",
            ]

            for process in processes:
                try:
                    # Find process IDs and kill them
                    kill_command = f"ps -aux | grep {process} | grep -v grep | awk '{{print $2}}' | xargs kill"
                    while kill_command:
                        _ = subprocess.run(
                            kill_command,
                            shell=True,
                            capture_output=True,
                            text=True,
                            check=True,
                        )
                        kill_command = f"ps -aux | grep {process} | grep -v grep | awk '{{print $2}}' | xargs kill"
                except Exception as e:
                    console.print(f"[bold red]{process} already dead[/]")

        jit_print(">> Hard kill finished")


def shut_down(settings: JitSettings, name, pid):
    console.print(f"Shutting down {name} with PID {pid}")
    if pid:
        try:
            # Terminate the process
            os.kill(pid, signal.SIGTERM)
        except ProcessLookupError:
            print(f"Process with PID {pid} does not exist.")
        except PermissionError:
            print(f"Permission denied to kill process with PID {pid}.")
        except Exception as e:
            print(f"An error occurred: {e}")


def log_dir(settings: JitSettings):
    if not settings.log_dir:
        # Define default LOG_DIR if not set
        settings.log_dir = f"logs_nodes{settings.nodes}_Jobid{settings.job_id}"
        # settings.gkfs_mntdir = f"{settings.gkfs_mntdir}_Jobid{settings.job_id}"
        # settings.gkfs_rootdir = f"{settings.gkfs_rootdir}_Jobid{settings.job_id}"

        if settings.log_suffix:
            settings.log_dir += f"_{settings.log_suffix}"

    counter = 0
    name = settings.log_dir
    while os.path.exists(settings.log_dir):
        counter += 1
        settings.log_dir = f"{name}_{counter}"

    # Resolve and return the absolute path of LOG_DIR
    settings.log_dir = os.path.abspath(settings.log_dir)

    # Create directory if it does not exist
    os.makedirs(settings.log_dir, exist_ok=True)

    settings.set_log_dirs()


def get_address_ftio(settings: JitSettings) -> None:
    # Get Address and port
    jit_print(">> Getting FTIO ADDRESS")
    if settings.cluster:
        call = f"srun --jobid={settings.job_id} {settings.ftio_node_command} --disable-status -N 1 --ntasks=1 --cpus-per-task=1 --ntasks-per-node=1 --overcommit --overlap --oversubscribe --mem=0 ip addr | grep ib0 | awk '{{print $2}}' | cut -d'/' -f1 | tail -1"
        jit_print(f"[bold cyan]>> Executing: {call}")
        try:
            result = subprocess.run(
                call, shell=True, capture_output=True, text=True, check=True
            )
            settings.address_ftio = result.stdout.strip()
        except subprocess.CalledProcessError as e:
            jit_print(f"[bold red]>>Error occurred: {e}")
            settings.address_ftio = ""

    jit_print(f">> Address FTIO: {settings.address_ftio}", True)


def get_address_cargo(settings: JitSettings) -> None:
    jit_print(">> Getting Cargo ADDRESS")
    if settings.cluster:
        call = f"srun --jobid={settings.job_id} {settings.single_node_command} --disable-status -N 1 --ntasks=1 --cpus-per-task=1 --ntasks-per-node=1 --overcommit --overlap --oversubscribe --mem=0 ip addr | grep ib0 | awk '{{print $2}}' | cut -d'/' -f1 | tail -1"
        jit_print(f"[bold cyan]>> Executing: {call}")
        try:
            result = subprocess.run(
                call, shell=True, capture_output=True, text=True, check=True
            )
            settings.address_cargo = result.stdout.strip()
            settings.cargo_server = (
                f"{settings.gkfs_daemon_protocol}://{settings.address_cargo}:62000"
            )
        except subprocess.CalledProcessError as e:
            jit_print(f"[bold red]>>Error occurred: {e}")
            settings.address_cargo = ""
            settings.cargo_server = ""
    else:
        settings.cargo_server = f"ofi+tcp://{settings.address_cargo}:62000"

    jit_print(f">> Address CARGO: {settings.address_cargo}")
    jit_print(f">> CARGO server:  {settings.cargo_server} \n")


def set_dir_gekko(settings: JitSettings) -> None:
    if settings.node_local and settings.cluster:
        jit_print(">> Node local flag set")
        # old_gkfs_rootdir = settings.gkfs_rootdir
        old_gkfs_mntdir  = settings.gkfs_mntdir
        
        settings.gkfs_rootdir = (
            f"/localscratch/{settings.job_id}/{os.path.basename(settings.gkfs_rootdir)}"
        )
        settings.gkfs_mntdir = (
            f"/localscratch/{settings.job_id}/{os.path.basename(settings.gkfs_mntdir)}"
        )
        jit_print(f">> |-> Gekko root dir updated to: {settings.gkfs_rootdir}")
        jit_print(f">> |-> Gekko mnt dir updated to: {settings.gkfs_mntdir}")
        
        if old_gkfs_mntdir in settings.run_dir:
            settings.run_dir = settings.run_dir.replace(old_gkfs_mntdir, settings.gkfs_mntdir)
            jit_print(
            f">> |-> Run dir updated to: {settings.run_dir}",
        )
        if old_gkfs_mntdir in settings.app_flags:
            settings.app_flags = settings.app_flags.replace(old_gkfs_mntdir, settings.gkfs_mntdir)
            jit_print(
            f">> |-> App flags updated to: {settings.app_flags}",
        )
            


def print_settings(settings) -> None:
    # Default values
    ftio_status = f"[bold green]ON[/]"
    gkfs_daemon_status = f"[bold green]ON[/]"
    gkfs_proxy_status = f"[bold green]ON[/]"
    cargo_status = f"[bold green]ON[/]"

    task_daemon = f"{settings.app_nodes}"
    cpu_daemon = f"{settings.procs_daemon}"
    task_proxy = f"{settings.app_nodes}"
    cpu_proxy = f"{settings.procs_proxy}"
    task_cargo = f"{settings.app_nodes*settings.procs_cargo}"
    cpu_cargo = f"{settings.procs_cargo}"
    task_ftio = "1"
    cpu_ftio = f"{settings.procs_ftio}"

    # Default settings text
    ftio_text = f"""
├─ ftio location  : {settings.ftio_bin_location}
├─ address ftio   : {settings.address_ftio}
├─ port           : {settings.port}
├─ # nodes        : 1
└─ ftio node      : {settings.ftio_node_command.replace('--nodelist=', '')}"""

    gkfs_daemon_text = f"""
├─ gkfs daemon    : {settings.gkfs_daemon}
├─ gkfs intercept : {settings.gkfs_intercept}
├─ gkfs mntdir    : {settings.gkfs_mntdir}
├─ gkfs rootdir   : {settings.gkfs_rootdir}
├─ gkfs hostfile  : {settings.gkfs_hostfile}"""

    gkfs_proxy_text = f"""
├─ gkfs proxy     : {settings.gkfs_proxy}
└─ gkfs proxyfile : {settings.gkfs_proxyfile}"""

    cargo_text = f"""
├─ cargo          : {settings.cargo}
├─ cargo cli      : {settings.cargo_bin}
├─ stage in path  : {settings.stage_in_path}
└─ address cargo  : {settings.address_cargo}"""

    if settings.exclude_ftio:
        ftio_text = """
├─ ftio activate  : [yellow]none[/]
├─ address ftio   : [yellow]none[/]
├─ port           : [yellow]none[/]
├─ # nodes        : [yellow]none[/]
└─ ftio node      : [yellow]none[/]"""
        ftio_status = "[bold yellow]off[/]"
        task_ftio = "[bold yellow]-[/]"
        cpu_ftio = "[bold yellow]-[/]"

    if settings.exclude_daemon:
        gkfs_daemon_text = """
├─ gkfs daemon    : [yellow]none[/]
├─ gkfs intercept : [yellow]none[/]
├─ gkfs mntdir    : [yellow]none[/]
├─ gkfs rootdir   : [yellow]none[/]
├─ gkfs hostfile  : [yellow]none[/]"""
        gkfs_daemon_status = "[bold yellow]off[/]"
        task_daemon = "[bold yellow]-[/]"
        cpu_daemon = "[bold yellow]-[/]"

    if settings.exclude_proxy:
        gkfs_proxy_text = """
├─ gkfs proxy     : [yellow]none[/]
└─ gkfs proxyfile : [yellow]none[/]"""
        gkfs_proxy_status = "[bold yellow]off[/]"
        task_proxy = "[bold yellow]-[/]"
        cpu_proxy = "[bold yellow]-[/]"

    if settings.exclude_cargo:
        cargo_text = """
├─ cargo location : [yellow]none[/]
├─ cargo cli      : [yellow]none[/]
├─ stage in path  : [yellow]none[/]
└─ address cargo  : [yellow]none[/]"""
        cargo_status = "[bold yellow]off[/]"
        task_cargo = "[bold yellow]-[/]"
        cpu_cargo = "[bold yellow]-[/]"

    # print settings
    text = f"""
[bold green]Settings
##################[/]
[bold green]setup[/]
├─ logs dir       : {settings.log_dir}
├─ pwd            : {settings.dir}
├─ ftio           : {ftio_status}
├─ gkfs daemon    : {gkfs_daemon_status}
├─ gkfs proxy     : {gkfs_proxy_status}
├─ cargo          : {cargo_status}
├─ cluster        : {settings.cluster}
├─ total nodes    : {settings.nodes}
|   ├─ app        : {settings.app_nodes}
|   └─ ftio       : 1
├─ tasks per node : -  
|   ├─ app        : {settings.procs_app} 
|   ├─ daemon     : {task_daemon}
|   ├─ proxy      : {task_proxy}
|   ├─ cargo      : {task_cargo}
|   └─ ftio       : {task_ftio}
├─ cpus per task  : {settings.procs} 
|   ├─ app        : 1
|   ├─ daemon     : {cpu_daemon}
|   ├─ proxy      : {cpu_proxy}
|   ├─ cargo      : {cpu_cargo}
|   └─ ftio       : {cpu_ftio}
├─ OMP threads    : {settings.omp_threads}
├─ max time       : {settings.max_time}
└─ job id         : {settings.job_id}

[bold green]ftio[/]{ftio_text}

[bold green]gekko[/]{gkfs_daemon_text}{gkfs_proxy_text}

[bold green] cargo[/]{cargo_text}

[bold green]app[/]
├─ run dir        : {settings.run_dir}
├─ app call       : {settings.app_call}
├─ # nodes        : {settings.app_nodes}
└─ app nodes      : {settings.app_nodes_command.replace('--nodelist=', '')}
[bold green]##################[/]
"""
    console.print(text)
    print_to_file(text, os.path.join(settings.log_dir, "settings.log"))


def print_to_file(text, file):
    remove = ["bold", "green", "yellow", "red", "cyan", "[/]", "[ ]", "[]"]
    for r in remove:
        text = text.replace(r, "")

    with open(file, "a") as file:
        file.write(text)


def jit_print(s: str, new_line: bool = False) -> None:
    if new_line:
        console.print(f"\n[bold green]JIT[/][green] {s}[/]")
    else:
        console.print(f"[bold green]JIT[/][green] {s}[/]")


def create_hostfile(settings):
    jit_print(f"[cyan]>> Cleaning Hostfile: {settings.gkfs_hostfile}")

    try:
        if os.path.exists(settings.gkfs_hostfile):
            os.remove(settings.gkfs_hostfile)
            jit_print("[yellow]>> Hostfile removed[/]")
        else:
            jit_print("[green]>> No hostfile found to remove[/]")

    except Exception as e:
        jit_print(f"[bold red]Error removing hostfile:[/bold red] {e}", True)

    # # Optionally, recreate the hostfile and populate it if needed
    # try:
    #     # Create the hostfile (if needed)
    #     # with open(settings.gkfs_hostfile, 'w') as file:
    #     #     for node in settings.nodes_arr[:-1]:  # Exclude the last element
    #     #         file.write(f"cpu{node}\n")
    #     # console.print(f"[cyan]>> Creating Hostfile: {settings.gkfs_hostfile}")
    #     pass
    # except Exception as e:
    #     console.print(f"[bold red]Error creating hostfile:[/bold red] {e}")


def format_time(elapsed):
    """Format the elapsed in a more readable way."""
    hours, remainder = divmod(elapsed, 3600)
    minutes, seconds = divmod(remainder, 60)
    return f"{int(hours)}h {int(minutes)}m {seconds:.2f}s"


def elapsed_time(settings: JitSettings, runtime: JitTime, name, elapsed):
    """Calculate and print the elapsed time."""

    elapsed_formatted = format_time(elapsed)
    log_message = (
        f"\n\n[cyan]############[JIT]##############\n"
        f"# {name}\n"
        f"# time: [yellow]{elapsed_formatted} [cyan]\n"
        f"# [yellow]{elapsed} [cyan]seconds\n"
        f"##############################\n\n"
    )
    console.print(log_message)

    # Write to log file
    print_to_file(
        log_message,
        os.path.join(settings.log_dir, "time.log"),
    )

    if "stage" in name.lower():
        if "in" in name.lower():
            runtime.stage_in = elapsed
        elif "out" in name.lower():
            runtime.stage_out = elapsed
    elif "app" in name.lower():
        runtime.app = elapsed


def check(settings: JitSettings):
    if settings.dry_run or settings.exclude_daemon:
        return

    call = flaged_mpiexec_call(settings, f"ls -lahrt {settings.gkfs_mntdir}")
    try:

        files = subprocess.check_output(
            f"{call}",
            shell=True,
            text=True,
        )
        jit_print(f"[cyan]>> {call}: [/]\n{files}")
    except subprocess.CalledProcessError as e:
        jit_print(f"[red]>> Failed to list files in {settings.gkfs_mntdir}: {e}[/]")


def flaged_mpiexec_call(settings: JitSettings, call: str, procs: int = 1) -> str:
    additional_arguments = load_flags_mpiexec(settings)
    call, procs = clean_call(call,procs)
    if settings.cluster:
        call = mpiexec_call(settings, call, procs, additional_arguments)
    else:
        call = (
            f"mpiexec -np {procs} --oversubscribe "
            # f"-x LD_LIBRARY_PATH={os.environ.get('LD_LIBRARY_PATH')} "
            f"{additional_arguments} {call}"
        )

    return call


def load_flags_mpiexec(settings: JitSettings, ftio_metrics: bool = False) -> str:
    additional_arguments = ""
    if not settings.exclude_ftio and ftio_metrics:
        additional_arguments += f"-x LIBGKFS_METRICS_IP_PORT={settings.address_ftio}:{settings.port} -x LIBGKFS_ENABLE_METRICS=on "
    if not settings.exclude_proxy:
        additional_arguments += f"-x LIBGKFS_PROXY_PID_FILE={settings.gkfs_proxyfile} "
    if not settings.exclude_daemon:
        additional_arguments += (
            f"-x LIBGKFS_LOG=info,warnings,errors "
            f"-x LIBGKFS_LOG_OUTPUT={settings.gekko_client_log} "
            f"-x LIBGKFS_HOSTS_FILE={settings.gkfs_hostfile} "
            f"-x LD_PRELOAD={settings.gkfs_intercept} "
        )
    additional_arguments += get_env(settings,"mpi")
    return additional_arguments


def load_flags(settings: JitSettings, ftio_metrics: bool = False) -> str:
    additional_arguments = ""
    if not settings.exclude_ftio and ftio_metrics:
        additional_arguments += f" LIBGKFS_METRICS_IP_PORT={settings.address_ftio}:{settings.port}  LIBGKFS_ENABLE_METRICS=on "
    if not settings.exclude_proxy:
        additional_arguments += f" LIBGKFS_PROXY_PID_FILE={settings.gkfs_proxyfile} "
    if not settings.exclude_daemon:
        additional_arguments += (
            f" LIBGKFS_LOG=info,warnings,errors "
            f" LIBGKFS_LOG_OUTPUT={settings.gekko_client_log} "
            f" LIBGKFS_HOSTS_FILE={settings.gkfs_hostfile} "
            f" LD_PRELOAD={settings.gkfs_intercept} "
        )
    return additional_arguments


def mpiexec_call(
    settings: JitSettings, command: str, procs: int = 1, additional_arguments=""
) -> str:
    call = (
        f" mpiexec -np {procs} --oversubscribe "
        f"--hostfile {settings.dir}/hostfile_mpi -map-by node "
        f"{additional_arguments} "
        f"{settings.task_set_0} {command}"
    )
    return call


def clean_call(call:str,procs:int):
    if "mpiexec" in call or "mpirun" in call:
        call = call.replace("mpiexec", "").replace("mpirun", "").strip()
        parts = call.split()
        if '-np' in parts:
            procs = int(parts[parts.index('-np') + 1]) 
            parts = [part for part in parts if part != '-np' and part != str(procs)]
            call = " ".join(parts)
    else:
        pass

    return call, procs


def get_executable_realpath(executable_name, search_location=None):
    """
    Try to find the real path of an executable.     
    Parameters:
    - executable_name (str): The name of the executable.
    
    Returns:
    - str: Real path of the executable or its name if not found.
    """
    if search_location:
        potential_path = os.path.join(search_location, executable_name)
        if os.path.isfile(potential_path) and os.access(potential_path, os.X_OK):
            try:
                return os.path.realpath(potential_path)
            except Exception as e:
                print(f"Warning: Could not resolve real path for {potential_path}: {e}")
                return executable_name
    
    # Fall back to searching in the system PATH
    executable_path = shutil.which(executable_name)
    if executable_path:
        try:
            return os.path.realpath(executable_path)
        except Exception as e:
            print(f"Warning: Could not resolve real path for {executable_name}: {e}")
    
    # Fallback: return the name if not found
    jit_print(f">> Application: {executable_name}")
    return executable_name

def update_hostfile_mpi(settings: JitSettings) -> None:
    # Command to get the list of hostnames for the job
    squeue_command = f"squeue -j {settings.job_id} -o '%N' | tail -n +2"
    scontrol_command = f"scontrol show hostname $({squeue_command})"

    # Execute the command and capture the hostnames
    hostnames = (
        subprocess.check_output(scontrol_command, shell=True).decode().splitlines()
    )

    # Write the hostnames to the file, excluding the Ftio node
    with open(f"{settings.dir}/hostfile_mpi", "w") as hostfile:
        for hostname in hostnames:
            if hostname.strip() != settings.ftio_node:
                hostfile.write(hostname + "\n")


def log_failed_jobs(settings: JitSettings, info: str) -> None:
    """Add failed job to a file

    Args:
        settings (JitSettings): Jit settings
        info (str): log text
    """
    parent = os.path.dirname(settings.log_dir)
    execution_path = os.path.join(parent, "execution.txt")
    try:
        jit_print(
            f"[yellow]>> Adding execution to list of failed jobs in {execution_path}.[/]"
        )
        with open(execution_path, "a") as file:
            file.write(f"- {info}\n")
    except:
        jit_print(f"[Red]>> Killing Job: {info}.\n Exiting script.[/]")


def set_env(settings: JitSettings):
    for _, key in enumerate(settings.env_var):
        jit_print(f"[green]>> Setting {key} to {settings.env_var[key]}[/]")
        os.environ[str(key)] = str(settings.env_var[key])
        jit_print(f"[cyan]>> {key} set to {os.getenv(str(key))}[/]")
        if os.getenv(str(key)) is None:
            jit_print(
                f"[red bold]>> {key} not set, do this manually:\nexport {key}={settings.env_var[key]}[/]"
            )


def get_env(settings: JitSettings,mode="srun") -> str:
    env = ""
    if "mpi" in mode:
        env = " ".join(f"-x {key}={value}" for key, value in settings.env_var.items())
    elif "srun": #srun
        env = ",".join(f"{key}={value}" for key, value in settings.env_var.items())
        env = "," + env
    else:
        pass
    return env




def save_bandwidth(settings: JitSettings):
    if not settings.exclude_ftio:
        try:
            command = f"cp {os.path.dirname(settings.log_dir)}/bandwidth.json {settings.log_dir}/bandwidth.json || true"
            _ = subprocess.run(
                command, shell=True, capture_output=True, text=True, check=True
            )
        except Exception as e:
            jit_print(f"[red] >> Error saving bandwidth:\n{e}")



import os
import sys
import re
import json
import time
from rich.console import Console
from rich.progress import Progress, SpinnerColumn, TimeElapsedColumn
from rich.panel import Panel
import numpy as np
import pandas as pd
from ftio.api.trace_analysis.trace_ftio_v2 import main as trace_ftio

console = Console()


def main(argv=sys.argv[1:]) -> None:
    verbose = False
    name = "plafrim"
    # specify the name with -n
    if "-n" in argv:
        index = argv.index("-n")
        name = str(argv[index + 1])
        argv.pop(index)
        argv.pop(index)
    if "-v" in argv:
        index = argv.index("-v")
        verbose = bool(argv[index + 1])
        argv.pop(index)
        argv.pop(index)

    # print(argv)
    start_time = time.time()
    # pattern = "_signal_plafrim.csv"
    pattern = f"_signal_{name}.csv"
    df = pd.DataFrame()

    # Handle command-line arguments properly
    if len(argv) > 0:
        folder_path = argv[0]
        argv.pop(0)
    else:
        # folder_path = os.getcwd()
        folder_path = "/d/github/FTIO/ftio/api/trace_analysis/plafrim"

    folder_path = os.path.abspath(folder_path)
    console.print(f"[bold green]Path is: {folder_path}[/]")

    # Create a list to store all matching csv files
    csv_files = []

    # Walk through all subdirectories and find matching files
    for dirpath, _, filenames in os.walk(folder_path):
        for filename in filenames:
            if filename.endswith(pattern):
                csv_files.append(os.path.join(dirpath, filename))

    if not csv_files:
        console.print(f"[bold red]No files matched the pattern: {pattern}![/]")
        return

    progress = Progress(
        SpinnerColumn(),
        *Progress.get_default_columns(),
        "[yellow] --  elapsed time",
        TimeElapsedColumn(),
    )

    try:
        with progress:
            task = progress.add_task("[green]Processing files...", total=len(csv_files))

            # Iterate over each csv file
            for file_path in csv_files:
                # Display the file being processed
                progress.console.print(
                    f"Processing ({csv_files.index(file_path)}/{len(csv_files)}): {file_path}"
                )

                # Run the trace_ftio function
                res = trace_ftio([file_path] + argv, verbose)

                # Create the new file name by replacing the pattern
                base_name = os.path.basename(file_path)
                json_name = base_name.replace(
                    f"_signal_{name}.csv", f"_freq_{name}.json"
                )
                json_path = os.path.join(os.path.dirname(file_path), json_name)
                data_converted = convert_dict(res)

                if json_name.endswith("json"):
                    with open(json_path, "w") as file:
                        json.dump(data_converted, file, indent=4)
                else:
                    console.print(
                        f"[bold red]Cannot dump Json file in {json_path}[/]"
                    )

                flat_res = flatten_dict(res)
                try:
                    flat_res["job_id"] = base_name.split("_")[0]
                    flat_res["file"] = file_path
                except:
                    flat_res["job_id"] = "??"
                    flat_res["file"] = "??"
                    console.print("[bold red]Unable to extract job id/path[/]")

                new_row_df = pd.DataFrame([flat_res])
                # Append the new row DataFrame to the existing DataFrame
                df = pd.concat([df, new_row_df], ignore_index=True)

                # Update the progress bar
                progress.advance(task)

        progress.console.print("[bold green]All files processed successfully![/]\n")
        console.print(
            f"[blue]FTIO total time:[/] {time.time()  - start_time:.4f} seconds\n"
            f"[blue]Location:[/] {folder_path}\n"
            f"[blue]Pattern:[/] {pattern}\n"
        )
        statistics(df)
    except KeyboardInterrupt:
        progress.console.print("[bold red]Keyboard interrupt![/]\n")
        statistics(df)
        sys.exit()
    console.print(f"[blue]Execution time:[/] {time.time()  - start_time:.4f} seconds")


def convert_dict(data):
    """Convert NumPy arrays and sets to lists in the dictionary."""
    if isinstance(data, dict):
        for key, value in data.items():
            if isinstance(value, np.ndarray):
                data[key] = value.tolist()
            elif isinstance(value, set):
                data[key] = list(value)
            elif isinstance(value, dict):
                convert_dict(value)
    return data


def flatten_dict(d):
    """Flatten the dictionary for DataFrame insertion."""
    flat = {}
    for key, value in d.items():
        if isinstance(value, dict):
            for sub_key, sub_value in value.items():
                if isinstance(sub_value, np.ndarray):
                    # Convert numpy arrays to lists
                    flat[f"{key}_{sub_key}"] = sub_value.tolist()
                elif isinstance(sub_value, set):
                    # Convert sets to lists
                    flat[f"{key}_{sub_key}"] = list(sub_value)
                else:
                    flat[f"{key}_{sub_key}"] = sub_value
        else:
            flat[key] = value
    return flat


def statistics(df,elapsed_time="",settings={}) -> None:
    # print(df)
    df_dom = reduce_to_max_conf(df)
    prefixes = relevant_prefix(df)
    color = ["purple4", "gold3", "deep_sky_blue1"]
    path = settings["res_path"]
    content = ""
    with open(f"{path}/ftio_output.txt", 'w') as file:
        for prefix in prefixes:
            s = ""
            s += periodic_apps(df, prefix)
            s += compute_metrics(df_dom, prefix, "conf")
            s += compute_metrics(df_dom, prefix, "dominant_freq", "Hz")
            s += time_app(df, prefix)
            console.print(
                Panel.fit(
                    s,
                    title=prefix.capitalize(),
                    border_style=color[prefixes.index(prefix)],
                    title_align="left",
                )
            )
            console.print("\n")
            content += cleaned_text(f"{prefix.capitalize()}" + "\n----------------\n"+ s + "\n\n")
        file.write(content+cleaned_text(elapsed_time))
        if settings:
            for _, field in enumerate(settings):
                file.write(f"\n{field}: {settings[field]}")

    df.to_csv(f"{path}/ftio.csv", index=False)
    df_dom.to_csv(f"{path}/ftio_flat.csv", index=False)
    # print(dom_df)


def periodic_apps(df, prefix) -> str:
    all_freq = len(df[f"{prefix}_dominant_freq"])
    # n = df[f'{prefix}_dominant_freq'].apply(lambda x: len(x)>0).sum()
    n = df[f"{prefix}_dominant_freq"].apply(lambda x: not np.isnan(x)).sum()
    # out = f"[blue]Periodic {prefix.capitalize()}: {n:.3e}/{all_freq:.3e} ({n/all_freq*100:.3e}%)[/]"
    out = f"[blue]Periodic I/O:[/]\n - {n:.0f}/{all_freq:.0f} ({n/all_freq*100:.2f}%)\n\n"
    return out


def compute_metrics(df: pd.DataFrame, prefix, suffix="conf", unit="%", title="") -> str:
    min = np.nan
    max = np.nan
    mean = np.nan
    median = np.nan
    nanmean = np.nan
    nanmedian = np.nan
    if not title:
        title = suffix.capitalize()
    
    conf_col = f"{prefix}_{suffix}"
    if not df[conf_col].isna().all():
        min = np.min(df[conf_col])
        max = np.max(df[conf_col])
        mean = np.mean(df[conf_col])
        median = np.median(df[conf_col])
        nanmean = np.nanmean(df[conf_col])
        nanmedian = np.nanmedian(df[conf_col])
    
    scale = 100 if "conf" in suffix else 1
    # out = f"[green]{prefix.capitalize()} {title}:\n - range: [{min*scale:.3e},{max*scale:.3e}] {unit}\n - mean: {mean*scale:.3e} {unit}\n - nanmean: {nanmean*scale:.3e} {unit}\n - median: {median*scale:.3e} {unit}\n - nanmedian: {nanmedian*scale:.3e} {unit}\n[/]"
    out = f"[gray][green]{title}:[/]\n - range: [{min*scale:.3f},{max*scale:.3f}] {unit}\n - mean: {mean*scale:.3f} {unit}\n - nanmean: {nanmean*scale:.3f} {unit}\n - median: {median*scale:.3f} {unit}\n - nanmedian: {nanmedian*scale:.3f} {unit}\n\n[/]"
    return out


def time_app(df, prefix):
    df[f"{prefix}_time"] = df[f"{prefix}_t_end"] - df[f"{prefix}_t_start"]
    out = compute_metrics(df, prefix, "time", "sec", "I/O Time")
    return out


def reduce_to_max_conf(df: pd.DataFrame) -> pd.DataFrame:
    prefixes = relevant_prefix(df)
    print(prefixes)
    # Iterate over each row
    for i, row in df.iterrows():
        for prefix in prefixes:
            conf_col = f"{prefix}_conf"
            freq_col = f"{prefix}_dominant_freq"
            amp_col = f"{prefix}_amp"
            phi_col = f"{prefix}_phi"
            freq = np.nan
            conf = np.nan
            amp  = np.nan
            phi  = np.nan
            if isinstance(row[conf_col], list) and len(row[conf_col]) > 0:
                dominant_index = np.argmax(row[conf_col])
                freq = row[freq_col][dominant_index]
                conf = row[conf_col][dominant_index]
                amp  = row[amp_col][dominant_index]
                phi  = row[phi_col][dominant_index]
            df.at[i, freq_col] = freq
            df.at[i, conf_col] = conf
            df.at[i, amp_col]  = amp
            df.at[i, phi_col]  = phi

    return df

def relevant_prefix(df):
    prefixes = ["read", "write", "both"]
    res = []
    for mode in prefixes:
        matching_columns = df.columns[df.columns.str.contains(mode)]
        if len(matching_columns) > 0:
            res.append(mode)
    return res

def cleaned_text(text:str) -> str:
    # Remove color tags and placeholders for rich console
    text = re.sub(r'\[\w+\]', '', text)  # Remove [color] tags
    text = re.sub(r'\[\/\]', '', text)  # Remove [/] tags
    return text

if __name__ == "__main__":
    main(sys.argv[1:])



import os 
import sys
import numpy as np
from ftio.cli.ftio_core import core
from ftio.parse.args import parse_args
from ftio.freq._dft import display_prediction
from ftio.plot.freq_plot import convert_and_plot
from ftio.parse.bandwidth import overlap
from ftio.parse.csv_reader import read_csv_file


def main(argv=sys.argv):
    
    full_path = get_path(argv)
    arrays = read_csv_file(full_path)

    # Print the arrays
    for key, array in arrays.items():
        print(f"{key}: {array}")


    #get frequency from file name
    f_s = 0.1 #get this value from the name of a file
    t_s = 1/f_s

    ranks = 10
    b = np.array(arrays['read']).astype(float)
    t = np.arange(0,len(b)*t_s,t_s).astype(float)
    total_bytes = np.sum(np.repeat(t_s,len(b))*len(b))

    # command line arguments
    argv = ["-e", "no"] #["-e", "mat"]

    # set up data
    data = {
            "time": t,
            "bandwidth": b,
            "total_bytes": total_bytes,
            "ranks": ranks 
            }

    #parse args
    args = parse_args(argv,"ftio")

    # perform prediction
    prediction, dfs = core([data], args)


    # plot and print info
    convert_and_plot(data, dfs, args)
    display_prediction("ftio", prediction)


def get_path(argv):
    # Example usage
    if len(argv) > 1:
        path = argv[1]
        if os.path.isabs(path):
            # full path
            full_path = path
        else:
            # relative path
            full_path = f'{os.getcwd()}/{path}'
            print(f'current file: {full_path}')
    else:
        # provide the path manually 
        full_path = f'{os.getcwd()}/data/data.csv'

    print(f'current file: {full_path}')

    return full_path


if __name__ == "__main__":
    main(sys.argv)


# import json
import requests

class MetricProxy:
    def __init__(self, url="http://localhost:1337"):
        self.url = url

    def jobs(self) -> dict:
        return requests.get(f'{self.url}/profiles').json()

    def profile(self,jobid) -> dict:
        return requests.get(f'{self.url}/profiles/get?jobid={jobid}').json()

    def metric(self,jobid) -> dict:
        return requests.get(f'{self.url}/trace/metrics/?job={jobid}').json()

    def trace(self,jobid) -> dict:
        return requests.get(f'{self.url}/trace/json?jobid={jobid}').json()


    def trace_metric(self,jobid,metric) -> dict:
        return requests.get(f'{self.url}/trace/plot/?jobid={jobid}&filter={requests.utils.quote(metric)}').json()

    def jsonl(self, jobid) -> str:
        return requests.get(f'{self.url}/model/download?jobid={jobid}').text

    def models(self, jobid) -> dict:
        return requests.get(f'{self.url}/model/get?jobid={jobid}').json()

    def online_models(self, jobid) -> dict:
        """Get models during runtime

        Args:
            jobid (str): job id

        Returns:
            dict: json file
        """
        return requests.get(f'{self.url}/model/ftio?jobid={jobid}').json()

    def plotmodel(self, jobid, metric, end, start=0, step=1) :
        return requests.get(f'{self.url}/model/plot?jobid={jobid}&metric={requests.utils.quote(metric)}&start={start}&end={end}&step={step}').json()



# # Client OBJECT
# mp = MetricProxy()

# # Get a LIST of all JOBs
# jobs = mp.jobs()
# #print(json.dumps(jobs, indent=4))

# # Get a JSONL for this JOB
# id_of_first_job = jobs[0]["jobid"]
# jsonl = mp.jsonl(id_of_first_job)
# #print(jsonl)

# # Get the model formulas and associated errors
# models = mp.models(id_of_first_job)
# #print(json.dumps(models, indent=4))

# # Now get the profile of a given JOB
# profile = mp.profile(id_of_first_job)
# #print(json.dumps(profile, indent=4))

# #trace for cpu 
# proxy_cpu_load_average_percent =mp.trace_metric(id_of_first_job,"proxy_cpu_load_average_percent")

# first_counter_name = profile["counters"][0]["name"]

# plot = mp.plotmodel(id_of_first_job, first_counter_name, 64)
# print(json.dumps(plot, indent=4))




import sys
import os
from datetime import datetime
import json
import numpy as np
from ftio.cli.ftio_core import core
from ftio.parse.args import parse_args
from ftio.freq._dft import display_prediction
from ftio.plot.freq_plot import convert_and_plot
from ftio.parse.csv_reader import read_csv_file
# from ftio.api.trace_analysis.helper import quick_plot
from rich.console import Console
console = Console()



def main(argv=sys.argv[1:],verbose=True, json_flag=True):
    if json_flag:
        # Time step is provided in the data for the json
        return extract_arrays_from_json(argv,verbose)
    else:
        # CSV uses a implicit time step
        return extract_arrays_from_csv(argv,verbose)


def extract_arrays_from_json(argv=sys.argv[1:],verbose=True):
    full_path = get_path(argv,verbose)
    with open(full_path, 'r') as file:
        arrays = json.load(file)
    b_r = np.array([])
    t_r = np.array([])
    b_w = np.array([])
    t_w = np.array([])

    if 'read' in arrays:
        b_r = np.array(arrays['read']['bandwidth']['b_overlap_avr']).astype(float)
        t_r = np.array(arrays['read']['bandwidth']['t_overlap']).astype(float)
    if 'write' in arrays:
        b_w = np.array(arrays['write']['bandwidth']['b_overlap_avr']).astype(float)
        t_w = np.array(arrays['write']['bandwidth']['t_overlap']).astype(float)

    # adapt for FTIO
    # command line arguments
    argv = [x for x in argv if '.py' not in x and '.json' not in x]
    if not '-e' in argv:
        argv.extend(['-e', 'no'])

    if verbose:
        console.print(f"Args: {argv}")

    res = run_ftio_on_group(argv, verbose, arrays, b_r, t_r, b_w, t_w)
    return res


def extract_arrays_from_csv(argv=sys.argv[1:],verbose=True):
    full_path = get_path(argv,verbose)
    arrays = read_csv_file(full_path)
    b_r = np.array([])
    b_w = np.array([])
    b_b = np.array([])

    # for key, array in arrays.items():
    #     print(f"{key}: {array}")
    
    if 'read' in arrays:
        b_r = np.array(arrays['read']).astype(float)
    if 'write' in arrays:
        b_w = np.array(arrays['write']).astype(float)
    if 'both' in arrays:
        b_b = np.array(arrays['both']).astype(float)

    if 'timestamp' in arrays:
        entries = [datetime.strptime(ts, '%Y-%m-%d %H:%M:%S.%f') for ts in arrays['timestamp']]
        t_s = entries[0]
        time_diffs_in_seconds = [(dt - t_s).total_seconds() for dt in entries]
        t = np.array(time_diffs_in_seconds)
    else:
        t_step =np.nan
        if '--time-step' not in argv:
            t_step = 1
            if verbose:
                console.print(f"[bold]Sampling rate set to {t_step}[/]")
        else:
            flag_index = argv.index('--time-step')
            t_step = float(argv[flag_index + 1])
            argv[flag_index:flag_index+2] = []
            if verbose:
                console.print(f"[bold green]Sampling rate set to {t_step} sec[/]")

        t = np.arange(0,len(b_w)*t_step,t_step).astype(float)

    # set sampling frequency if not set
    if '-f' not in argv:
        argv.extend(['-f', f'{2/t_step}'])
        if verbose:
            console.print(f"[bold green]Sampling rate set to {t_step} sec ({2/t_step:.3f}) Hz[/]")
    # plot
    # quick_plot(t,b_w)

    # adapt for FTIO
    # command line arguments
    argv = [x for x in argv if '.py' not in x and '.csv' not in x]
    if not '-e' in argv:
        argv.extend(['-e', 'no'])

    if verbose:
        console.print(f"Args: {argv}")
    # argv = ['-e', 'mat']

    res = run_ftio_on_group(argv, verbose, arrays, b_r, t, b_w, t, b_b, t)
    return res


def run_ftio_on_group(argv, verbose, arrays, b_r, t_r, b_w, t_w, b_b=[], t_b=[]):
    total_bytes_r = 0#np.sum(np.repeat(t_s,len(b_r))*len(b_r))
    total_bytes_w = 0#np.sum(np.repeat(t_s,len(b_w))*len(b_w))
    total_bytes_b = 0#np.sum(np.repeat(t_s,len(b_b))*len(b_b))
    res_r={}
    res_w={}
    res_b={}
    
    if 'read' in arrays and len(b_r) > 1:
        res_r = quick_ftio(argv,b_r,t_r, total_bytes_r, 10, 'read',verbose)
    if 'write' in arrays and len(b_w) > 1:
        res_w = quick_ftio(argv,b_w,t_w, total_bytes_w, 10, 'write',verbose)
    if 'both' in arrays and len(b_b) > 1:
        res_b = quick_ftio(argv,b_b,t_b, total_bytes_b, 10, 'both',verbose)

    return {"read":res_r,"write":res_w,"both":res_b}


def get_path(argv, verbose=True):

    full_path =''
    # Example usage

    # Check if argv is a string
    if argv and isinstance(argv, str):
        argv = [argv]  # Convert string to a list with one element

    for i in argv:
        if 'csv' in i or 'json' in i:
            path = i
            if os.path.isabs(path):
                # full path
                full_path = path
            else:
                # relative path
                full_path = f'{os.getcwd()}/{path}'
            break

    if not full_path:
        full_path = f'{os.getcwd()}/data_2.csv'

    if verbose:
        console.print(f'[green] current file: {full_path}[/]')

    return full_path


def quick_ftio(argv,b,t, total_bytes, ranks, msg, verbose=True) :

    # set up data
    data = {
            'time': t,
            'bandwidth': b,
            'total_bytes': total_bytes,
            'ranks': ranks 
            }

    #parse args
    args = parse_args(argv,'ftio')

    # perform prediction
    prediction, dfs = core([data], args)


    # plot and print info
    convert_and_plot(data, dfs, args)
    if verbose:
        console.print(f'[green]>> Prediction for {msg}[/]')
        display_prediction("ftio", prediction)
    return prediction

if __name__ == "__main__":
    _ = main(sys.argv)


import argparse
import json
import os
import plotly.graph_objects as go
import numpy as np
from ftio.plot.helper import format_plot_and_ticks




def main(filenames):
    if not filenames:
        results = JitResult()
        # # run with x nodes  128 procs [jit | jit_no_ftio | pure] (now in old folder)
        # ################ Edit area ############################
        # title = "Nek5000 with 128 procs checkpointing every 10 steps with a total of 100 steps"
        # tmp_app = [207.3, 181.44, 181.56]
        # tmp_stage_out = [3.95,  17.68, 0]
        # tmp_stage_in =  [0.72,  0.75, 0]
        # results.add_experiment(tmp_app,tmp_stage_out,tmp_stage_in,"# 2")

        # # run with 3 nodes
        # tmp_app = [90.11, 84.81, 103]
        # tmp_stage_out = [2.26, 61.0, 0]
        # tmp_stage_in = [1.11, 1.11, 0]
        # results.add_experiment(tmp_app,tmp_stage_out,tmp_stage_in,"# 3")

        # tmp_app = [70.53, 72.71, 80.21]
        # tmp_stage_out = [1.891,9.430, 0]
        # tmp_stage_in = [1.149, 1.145, 0]
        # results.add_experiment(tmp_app,tmp_stage_out,tmp_stage_in,"# 4")

        # run with x nodes 32 procs [jit | jit_no_ftio | pure] (now in old folder)
        ################ Edit area ############################
        # # run with 3 nodes
        # tmp_app       = [ 157.8 , 156.51 , 160 ]
        # tmp_stage_out = [ 1.041 , 1.04 ,  0]
        # tmp_stage_in  = [ 1.099 ,  1.09,  0]
        # add_experiment(data,tmp_app,tmp_stage_out,tmp_stage_in,"# 3")
        

        # tmp_app       = [ 114.72,122.07  ,  130.95]
        # tmp_stage_out = [ 1.04, 1.03 ,  0]
        # tmp_stage_in  = [ 1.11, 1.83  ,  0]
        # add_experiment(data,tmp_app,tmp_stage_out,tmp_stage_in,"# 4")

        # tmp_app       = [ 97.11,106.31, 98.11]
        # tmp_stage_out = [ 1.05,1.12,0]
        # tmp_stage_in  = [ 1.13,1.12,0]
        # add_experiment(data,tmp_app,tmp_stage_out,tmp_stage_in,"# 5")


        # tmp_app       = [ 126.93,119.06, 93.27]
        # tmp_stage_out = [ 1.12,1.159,0]
        # tmp_stage_in  = [ 1.59,1.96,0]
        # add_experiment(data,tmp_app,tmp_stage_out,tmp_stage_in,"# 10")

        # tmp_app       = [ 182.58,174.67,90.64 ]
        # tmp_stage_out = [ 1.08,1.08,0]
        # tmp_stage_in  = [ 3.57,2.84,0]
        # add_experiment(data,tmp_app,tmp_stage_out,tmp_stage_in,"# 20")
        
        # title = "Nek5000 with 16 procs checkpointing every 10 steps with a total of 50 steps"
        # filename = "results_mogon/procs16_steps50_writeinterval10.json "
        
        title = "Nek5000 with 16 procs checkpointing every 5 steps with a total of 50 steps"
        filename = "results_mogon/wacom++_app_proc_1_OMPthreads_64_12500000.json"
        current_directory =  os.path.dirname(os.path.abspath(__file__))
        json_file_path = os.path.join(current_directory, filename)
        
        extract_and_plot(results,json_file_path,title)
    else:
        for filename in filenames:
            print(f"Processing file: {filename}")
            results = JitResult()
            title = filename
            current_directory = os.getcwd()
            json_file_path = os.path.join(current_directory, filename)
            extract_and_plot(results,json_file_path,title)
        

def extract_and_plot(results,json_file_path:str, title:str):
    with open(json_file_path, "r") as json_file:
        data = json.load(json_file)
        data = sorted(data, key=lambda x: x['nodes'])
        for d in data:
            results.add_dict(d)
    
    results.plot(title)


def add_mode(list1, list2):
    # Create a new list to store the result
    result = []
    step = int(len(list1)/3)
    
    # Iterate over the range of the lists' lengths
    for i in range(len(list2)):
        # Add the current element from each list    
        for j in range(step):
            result.append(list1[i*step+j])
        result.append(list2[i])
    
    return result



class JitResult:
    def __init__(self) -> None:
        self.app       = []
        self.stage_out = []
        self.stage_in  = []
        self.node = []

    def add_experiment(self,tmp_app:list[float],tmp_stage_out:list[float],tmp_stage_in:list[float],run:str):
        self.app = add_mode(self.app,tmp_app)
        self.stage_out = add_mode(self.stage_out,tmp_stage_out)
        self.stage_in = add_mode(self.stage_in,tmp_stage_in)
        self.node.append(run)


    def add_dict(self, data_list:list[dict]):
        tmp_app       = [0.0,0.0,0.0]
        tmp_stage_out = [0.0,0.0,0.0]
        tmp_stage_in  = [0.0,0.0,0.0]

        for i in data_list["data"]:
            if i["mode"] in {"DPCF","DCF"}:
                index = 0
            elif i["mode"] in {"DC","DPC"}:
                index = 1
            else:
                index = 2
            tmp_app[index] = i["app"]
            tmp_stage_out[index] = i["stage_out"]
            tmp_stage_in[index] = i["stage_in"]
                
        self.add_experiment(tmp_app,tmp_stage_out,tmp_stage_in,f"# {data_list["nodes"]}")
        

    def plot(self,title=""):
        # Sample data for the stacked plot
        categories = ['JIT', 'JIT no FTIO', 'Pure']
        repeated_strings = [s for s in categories for _ in self.node]
        repeated_numbers = self.node * len(categories)
        categories = [repeated_strings, repeated_numbers]

        fig = go.Figure()
        

        fig.add_bar(x=categories, y=self.app, text=self.app,  name="App")
        fig.add_bar(x=categories, y=self.stage_out, text=self.stage_out,  name="Stage out")
        fig.add_bar(x=categories, y=self.stage_in, text=self.stage_in,  name="Stage in")

        # text 
        # Update the layout to stack the bars
        fig.update_traces(textposition='inside', texttemplate = "%{text:.2f}", textfont_size=14, textangle=0, textfont=dict(
            # family="sans serif",
            color="white"
        ))
        # comment out to see all text
        fig.update_layout(uniformtext_minsize=8, uniformtext_mode='hide')
        # sum total
        total = list(np.round(np.array(self.app) + np.array(self.stage_out) + np.array(self.stage_in),2))
        # plot total
        fig.add_trace(go.Scatter(
            x=categories, 
            y=total,
            text=total,
            mode='text',
            textposition='top center',
            textfont=dict(
                size=14,
            ),
            showlegend=False
        ))

        
        fig.update_layout(
            yaxis_title="Time (s)",
            xaxis_title=f"Experimental Runs with # Nodes",
            showlegend=True,
            title=title,
            barmode="relative",
            width=1500,
            height=600,
            legend=dict(
                orientation="h",
                yanchor="bottom",
                y=0.9,
                xanchor="right",
                x=0.995
                )
            )

        
        format_plot_and_ticks(fig, x_minor=False)
        # Display the plot
        fig.show()




if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Load JSON data from files and plot.")
    parser.add_argument(
        'filenames',
        type=str,
        nargs="*",  # '*' allows zero or more filenames
        default=[],
        help="The paths to the JSON file(s) to plot."
    )
    args = parser.parse_args()
    
    main(args.filenames)
    


import sys
import plotly.express as px
import pandas as pd
import plotly.graph_objects as go
import numpy as np 

from ftio.plot.plot_core import PlotCore
from ftio.plot.helper import legend_fix
from ftio.plot.helper import format_plot

def main(args=sys.argv):
    plotter = PlotCore(args)

    colors = px.colors.qualitative.Plotly
    plotter.nRun = len(pd.unique(plotter.data.df_time["file_index"]))
    symbols = ["circle", "square", "cross", "star-triangle-down", "hourglass"]
    markeredgecolor = "DarkSlateGrey"


    if plotter.nRun == 1:
        x = plotter.data.df_time["number_of_ranks"].astype(str)
        
    else:
        plotter.data.df_time = plotter.data.df_time.sort_values(
            by=["number_of_ranks", "file_index"]
        )
        x = [
            plotter.data.df_time["number_of_ranks"].astype(str),
            plotter.data.df_time["file_index"].astype(str),
        ]

    fig = go.Figure()
    ranks = plotter.data.df_time["number_of_ranks"]
    # sets_names = ["Limit", "No Limit"]
    # sets = [[0,1],[2]]
    sets = [[0,1],[2,3],[4,5],[6,7]]
    sets_names = ["Direct","Up-only","Adaptive","No Limit"]

    metric = [plotter.data.df_time["delta_t_agg"],plotter.data.df_time["delta_t_agg_io"]]
    modes = ["Total", "IO"]

    for i,_ in enumerate (sets):
        group = plotter.data.df_time["file_index"].isin(values=sets[i])
        x = ranks[group]
        x_unqiue = pd.unique(x)
        y = np.zeros((len(modes), len(x_unqiue)))
        y_plus = np.zeros((len(modes), len(x_unqiue)))
        y_minus = np.zeros((len(modes), len(x_unqiue)))

        #fill the array
        for count0, _ in enumerate(modes):
            for count1, _ in enumerate(x_unqiue):
                scope_metric = metric[count0]
                scope_metric = scope_metric[group]
                data = scope_metric[x == x_unqiue[count1]]
                y[count0, count1] = data.mean()
                y_plus[count0, count1] = abs(max(data - y[count0, count1]))
                y_minus[count0, count1] = abs(min(data - y[count0, count1]))
            
            fig.add_trace(go.Scatter(
                    x=x_unqiue,
                    y=y[count0],
                    error_y=dict(
                        type="data",
                        symmetric=False,
                        array=y_plus[count0],
                        arrayminus=y_minus[count0],
                    ),
                    mode="lines+markers",
                    name=modes[count0],
                    legendgroup=sets_names[i],
                    legendgrouptitle_text=sets_names[i],
                    marker=dict(
                        symbol=symbols[i], line=dict(width=1, color=markeredgecolor)
                    ),
                    # marker_color=colors[i]
                    showlegend=True
                ))

    fig.update_xaxes(title_text="Ranks")
    fig.update_layout(xaxis_title="Ranks",yaxis_title="Aggregated Time (s)")
    fig.update_layout(legend_groupclick= 'toggleitem')
    fig = format_plot(fig)
    fig.update_layout(
        hovermode="x",
        legend_tracegroupgap=1,
        width=1200,
        height=500,
    )
    fig.show()
    print(plotter.names)

if __name__ == "__main__":
    main(sys.argv)


import numpy as np


class PhaseMode:
    def __init__(self, name: str, matches: list[str]) -> None:
        self.name = name
        self.matches = matches
        self.data = []
        self.wave = np.nan
        self.t    = np.array([])

    def get(self, attribute: str):
        if hasattr(self, attribute):
            return getattr(self, attribute)
        else:
            raise AttributeError(f"'MyClass' object has no attribute '{attribute}'")


    def match(self, d: dict):
        if any(n in d['metric'] for n in self.matches):
            return True
        else:
            return False

    def add(self, d: dict) -> None:
        self.data.append(d)

    def set_time(self,t:np.ndarray) -> None:
        self.t = t

    def aggregates_waves(self, t, normed=True):
        self.wave = np.zeros(len(t))
        if self.t.size != 0:
            for i in self.data:
                self.wave = self.wave + calculate_wave(i, self.t)[0]

            if normed:
                self.wave = norm(self.wave)

    def get_wave(self,metric:str) :
        wave = np.array([])
        name = ""
        if self.t.size != 0:
            for d in self.data:
                if metric in d['metric']:
                    wave,name =  calculate_wave(d, self.t)

        return (wave,name)



def calculate_wave(prediction:dict, t:np.ndarray = np.array([]))  :# -> tuple[NDArray[Any] | Any, str | LiteralString]:# -> tuple[NDArray[Any] | Any, str | LiteralString]:# -> tuple[NDArray[Any] | Any, str | LiteralString]:
    wave = np.array([])
    name = ""
    if prediction:
        n = int(np.floor((prediction["t_end"] - prediction["t_start"]) * prediction["freq"]))
        if t.size == 0:
            t = np.arange(prediction["t_start"], prediction["t_end"], 1 / prediction["freq"])

        if 'top_freq' in prediction:
            wave = np.zeros_like(t)
            for j in range(0, len(prediction['top_freq']['freq'])):
                amp  = prediction['top_freq']['amp'][j]
                freq = prediction['top_freq']['freq'][j]
                phi  =  prediction['top_freq']['phi'][j]
                wave = wave +  2 / n *amp * np.cos(2 * np.pi * freq * t+ phi)
                if j < 3:
                    name += f"{2 / n *amp:.1e}*cos(2\u03C0*{freq:.2e}*t+{phi:.2e})"
                elif j == 4:
                    name += "..."
        else:
            max_conf_index = np.argmax(prediction['conf'])
            amp  = prediction['amp'][max_conf_index]
            freq = prediction['dominant_freq'][max_conf_index]
            phi  = prediction['phi'][max_conf_index]
            wave =  2 / n *amp * np.cos(2 * np.pi * freq * t + phi)
            name += f"{2 / n *amp:.1e}*cos(2\u03C0*{freq:.2e}*t+{phi:.2e})"
    return (wave,name)

def norm(wave: np.ndarray) -> np.ndarray:
    max_value = np.max(np.abs(wave))
    normed_wave = wave / max_value if max_value > 0 else wave
    return normed_wave



from ftio.cli.ftio_core import core
from ftio.parse.args import parse_args
from ftio.freq._dft import display_prediction
import numpy as np
# from ftio.prediction.helper import get_dominant
# from ftio.plot.freq_plot import convert_and_plot
from ftio.freq.helper import MyConsole

CONSOLE = MyConsole()
CONSOLE.set(True)

#TODO: extend this similar to ftio/prediction/analysis.py

def ftio_task(metric:str, arrays:np.ndarray, argv:list, ranks:int, show:bool=True) -> dict:
    """generate FTIO prediction from 2d np array

    Args:
        metric (str): Name of metric
        arrays (np.ndarray): 2D np array containing the bandwidth and time
        argv (list): List of args for ftio
        ranks (int): Number of ranks
        show (bool): show the prediction result

    Returns:
        dict: prediction from FTIO (see core from ftio.cli.ftio_core)
    """
    if len(arrays[0]) > 1:
        # set up data
        data = {"bandwidth": arrays[0], "time": arrays[1], "total_bytes": 0, "ranks": ranks}

        # parse args
        args = parse_args(argv, "ftio")

        # perform prediction
        prediction, dfs = core([data], args)

        # # plot and print info
        # convert_and_plot(data, dfs, args)
        if show:
            CONSOLE.info(f"\n[green underline]Metric: {metric}[/]")
            display_prediction("ftio", prediction)

        return prediction


def ftio_task_save(data, metric:str, arrays:np.ndarray, argv:list, ranks:int, show:bool=False) -> None:
    prediction = ftio_task(metric, arrays ,argv ,ranks, show)
    # freq = get_dominant(prediction) #just get a single dominant value
    if prediction:
        data.append({
            "metric": f"{metric}",
            "dominant_freq": prediction["dominant_freq"],
            "conf": prediction["conf"],
            "amp": prediction["amp"],
            "phi": prediction["phi"],
            "t_start": prediction["t_start"],
            "t_end": prediction["t_end"],
            "total_bytes": prediction["total_bytes"],
            "ranks": prediction["ranks"],
            "freq": prediction["freq"],
            **({"top_freq": prediction["top_freq"]} if "top_freq" in prediction else {})
            }
        )
    else:
        CONSOLE.info(f"\n[yellow underline]Warning: {metric} returned {prediction}[/]")



import pandas as pd
import argparse
import os  # To handle file path operations
import sys
import plotly.express as px
import plotly.graph_objects as go
import plotly.subplots as sp
import numpy as np

def main(argv=sys.argv[1:]):
    # Parse command-line arguments for file paths
    if not argv:
        argv=["/d/traces/traces_from_plafrim/ftio_flat.csv"]
        
    parser = argparse.ArgumentParser(description="Process some CSV files.")
    parser.add_argument(
        "file_paths", metavar="F", type=str, nargs="+", help="Paths to the CSV files",
        
    )
    args = parser.parse_args(argv)

    # List to store the DataFrames and filenames
    dataframes = []
    file_basenames = []

    # Read the CSV files into DataFrames and store their basenames
    for file_path in args.file_paths:
        df = pd.read_csv(file_path)
        basename = os.path.basename(file_path)  # Extract basename from file path
        basename = os.path.splitext(basename)[0]  # Remove extension
        df["Source File"] = basename  # Add a new column with the basename
        dataframes.append(df)
        file_basenames.append(basename)

    # Combine all DataFrames into a single DataFrame
    combined_df = pd.concat(dataframes, ignore_index=True)

    # Reshape the DataFrame to long format
    # Use "Source File" for the color coding and melt the DataFrame
    combined_df_long = pd.melt(
        combined_df, id_vars=["Source File"], var_name="Metric", value_name="Value"
    )

    # plot 
    # single_plot(combined_df_long)
    sub_plots(combined_df_long)


def single_plot(df):
    # Use Plotly Express to create a box plot with color by filename
    fig = px.box(
        df,
        x="Metric",  # Column name for x-axis
        y="Value",  # Column name for y-axis
        color="Source File",  # Use the filename column for color
        title="Combined Box Plots by Metric",
        labels={"Metric": "Metric", "Value": "Values"},
    )  # Labels for the axes

    # Update layout
    fig.update_layout(
        boxmode="group",  # Group the boxes for better comparison
        height=600,
        width=1200,
        xaxis_title="Metric",
        yaxis_title="Values",
    )

    # Show the plot
    fig.show()


def sub_plots(df):
    # Get unique metrics
    metrics = df["Metric"].unique()
    n = len(df["Source File"].unique())
    metrics = metrics[:-2] #remove jobid and file name
    num_metrics = len(metrics)

    order = []
    subplot_titles = []
    fields = ["read","write","both"]
    for i, field in enumerate(fields):
        row = 0
        for j,metric in enumerate(metrics):
            if field in metric:
                row  += 1 
                subplot_titles.append(metric)
                order.append([i+1,row])

    suffixes = []
    for title in subplot_titles:
        suffix = title.split('_', 1)[1]  # Split only at the first underscore
        if suffix not in suffixes:
            suffixes.append(suffix)

    # Reorder the list based on the prefix order and original suffix order
    subplot_titles = [
        f"{prefix}_{suffix}"
        for suffix in suffixes
        for prefix in fields
        if f"{prefix}_{suffix}" in subplot_titles
    ]

    col = 0
    for i in order:
        col = max(i[0],col)

    row = int(np.ceil(num_metrics/(col)))

    # Create subplots with each row representing a metric
    fig = sp.make_subplots(
        rows=row, cols=col, subplot_titles=subplot_titles
    )

    # Loop through each metric and add a box plot to the corresponding subplot
    for i, metric in enumerate(metrics):
        # Filter data for the current metric
        metric_df = df[df["Metric"] == metric]

        # Create a box plot for the current metric
        box_trace = go.Box(
            x=metric_df["Source File"], y=metric_df["Value"], name=metric,
            # hide points
            boxpoints=False, 
            # std and mean
            boxmean='sd'
        )
        # Clean the 'Value' column: Convert to numeric, coerce errors to NaN
        # metric_df['Value'] = pd.to_numeric(metric_df['Value'], errors='coerce')
        metric_df.loc[:, 'Value'] = pd.to_numeric(metric_df['Value'], errors='coerce')
        
        # Drop rows where 'Value' could not be converted to numeric (i.e., NaN)
        metric_df = metric_df.dropna(subset=['Value'])

        # Add the box plot to the appropriate subplot
        fig.add_trace(box_trace, row=order[i][1], col=order[i][0])
        for s in metric_df["Source File"].unique():
            fig.add_annotation(
                x=s,  # Use specific source file as x coordinate
                y=metric_df[metric_df["Source File"] == s]["Value"].max(),
                text=str(len(metric_df[metric_df["Source File"] == s]["Value"])),
                yshift=10,
                showarrow=False,
                col=order[i][0],
                row=order[i][1]
            )

    # Update layout for the entire figure
    fig.update_layout(
        title="Box Plots by Metric",
        height=400 * row,  # Adjust height for number of metrics
        width=300 +80*col*n,  # Adjust width for visibility
        xaxis_title="Source File",
        yaxis_title="Values",
        showlegend=False,  # Hide legend to reduce clutter
    )

    # Show the plot
    fig.show()
    print(f"Saving HTML filer out.html in {os.getcwd()}")
    fig.write_html(f"{os.getcwd()}/out.html")
    # print(
    #     f"Saving online HTML filer out_online.html in {os.getcwd()}."
    #     f"This file needs an internet connection to be viewed for plotly js"
    #     )
    # fig.write_html(f"{os.getcwd()}/out_online.html",include_plotlyjs=False)


if __name__ == "__main__":
    main(sys.argv[1:])



import os
import time
from datetime import datetime
from rich.console import Console
from ftio.api.gekkoFs.jit.execute_and_wait import execute_block
from ftio.api.gekkoFs.jit.jitsettings import JitSettings
from ftio.api.gekkoFs.jit.setup_helper import jit_print
console = Console()

def check_setup(settings:JitSettings):
    
    if not settings.exclude_all:

        # Display MPI hostfile
        if settings.cluster:
            mpi_hostfile_path = os.path.expanduser(f"{settings.dir}/hostfile_mpi")
            with open(mpi_hostfile_path, "r") as file:
                mpi_hostfile_content = file.read()
            console.print(f"[cyan]>> MPI hostfile:\n{mpi_hostfile_content}[/]")

        # Display GekkoFS hostfile
        gekkofs_hostfile = settings.gkfs_hostfile
        with open(gekkofs_hostfile, "r") as file:
            gekkofs_hostfile_content = file.read()
        console.print(f"[cyan]>> Geko hostfile:\n{gekkofs_hostfile_content}[/]")

        
        # ls_command = f"LD_PRELOAD={settings.gkfs_intercept} LIBGKFS_HOSTS_FILE={gekkofs_hostfile} ls {settings.gkfs_mntdir}"
        # files = subprocess.check_output(ls_command, shell=True).decode()
        # console.print(f"[cyan]>> geko_ls {gkfs_mntdir}: \n{files}[/]")

        if settings.cluster and settings.debug and not settings.exclude_daemon:
            #srun dies not work with gekko
            # call = (
            #     f"srun --jobid={settings.job_id} {settings.app_nodes_command} --disable-status -N {settings.app_nodes} --ntasks={settings.app_nodes} "
            #     f"--cpus-per-task={settings.procs_daemon} --ntasks-per-node=1 --overcommit --overlap --oversubscribe --mem=0 "
            #     f"--export=LIBGKFS_HOSTS_FILE={settings.gkfs_hostfile},LD_LIBRARY_PATH={os.environ.get('LD_LIBRARY_PATH')},LD_PRELOAD={settings.gkfs_intercept} "
            #     f"/usr/bin/ls {settings.gkfs_mntdir} "
            # )
            # jit_print("[cyan]>> Checking srun with Gekko")
            # out = execute_block(call, False)
            # console.print(f"srun check: {out}\n")


            additional_arguments = ""
            if not settings.exclude_ftio:
                additional_arguments += f"-x LIBGKFS_METRICS_IP_PORT={settings.address_ftio}:{settings.port} -x LIBGKFS_ENABLE_METRICS=on "
            if not settings.exclude_proxy:
                additional_arguments += f"-x LIBGKFS_PROXY_PID_FILE={settings.gkfs_proxyfile} "
            if not settings.exclude_daemon:
                additional_arguments += (
                    f"-x LIBGKFS_LOG=info,warnings,errors "
                    f"-x LIBGKFS_LOG_OUTPUT={settings.gekko_client_log} "
                    f"-x LIBGKFS_HOSTS_FILE={settings.gkfs_hostfile} "
                    f"-x LD_PRELOAD={settings.gkfs_intercept} "
                    )


            #test script
            jit_print("[cyan] >> Checking test file")
            try:
                timestamp = datetime.now().strftime('%Y_%m_%d_%H_%M_%S')
                file = create_test_file("test.sh"+timestamp, settings)
                call = (
                        f" mpiexec -np {settings.app_nodes} --oversubscribe "
                        f"--hostfile {settings.dir}/hostfile_mpi --map-by node "
                        f"{additional_arguments} "
                        
                        f"{file}"
                    )
                out = execute_block(call, False)
                console.print(f"{out}")
                # remove the created file
                if os.path.exists(file):
                    os.remove(file)
            except Exception as e:
                jit_print(f"[red] >> Error running test script:\n{e}")
        else:
            jit_print("[cyan]>> Skipping setup check")
        # # Run MPI exec test script
        # procs = settings.procs
        # if settings.cluster:
        #     test_script_command = (f"mpiexec -np {procs} --oversubscribe --hostfile {mpi_hostfile_path} "
        #                         f"--map-by node -x LIBGKFS_LOG=errors -x LD_PRELOAD={settings.gkfs_intercept} "
        #                         f"-x LIBGKFS_HOSTS_FILE={gekkofs_hostfile} -x LIBGKFS_PROXY_PID_FILE={settings.gkfs_proxyfile} "
        #                         f"/home/tarrafah/nhr-admire/tarraf/FTIO/ftio/api/gekkoFs/scripts/test.sh")
        #     console.print(f"[cyan]>> statx:[/]")
        #      subprocess.run(test_script_command, shell=True, text=True, capture_output=True, check=True, executable="/bin/bash"        )

                    
        #     srun_command = (f"srun --export=LIBGKFS_HOSTS_FILE={settings.gkfs_hostfile},LD_LIBRARY_PATH={os.environ.get('LD_LIBRARY_PATH')},"
        #                     f"LD_PRELOAD={settings.gkfs_intercept} --jobid={settings.job_id} {settings.app_nodes_command} --disable-status "
        #                     f"-N 1 --ntasks=1 --cpus-per-task=1 --ntasks-per-node=1 --overcommit --overlap --oversubscribe --mem=0 "
        #                     f"/usr/bin/ls {settings.gkfs_mntdir}")
        #     files2 = subprocess.check_output(srun_command, shell=True).decode()
        #     console.print(f"[cyan]>> srun ls {settings.gkfs_mntdir}: \n{files2}[/]")

        # Note: The commented out command for `mpirun` is not included in this translation.

    # Pause for 1 second
    time.sleep(1)


def create_test_file(name:str, settings:JitSettings) -> str:
    # Define the content of the shell script
    script_content = "#!/bin/bash\n"
    script_content +="myhostname=$(hostname)\n"
    script_content +=f"statcall=$(stat {settings.gkfs_mntdir})\n"
    if settings.app_nodes == 1:
        script_content +=f"lscall=$(ls -lR {settings.gkfs_mntdir})\n"
    else:
        script_content +=f"lscall=$(ls {settings.gkfs_mntdir})\n"
    script_content += """
    echo -e "Hello I am ${myhostname} and stat output: \\n${statcall}\\n The directory contains:\\n${lscall}\\n"
    """

    # Write the content to a file called tet.sh
    file_path = os.path.join(os.getcwd(), name)
    with open(file_path, "w") as file:
        file.write(script_content)

    os.chmod(file_path, 0o755)
    return file_path


import numpy as np
import pandas as pd
import plotly.express as px
import seaborn as sns
import matplotlib.pyplot as plt
import plotly.graph_objects as go
import fastcluster

from ftio.plot.helper import format_plot_and_ticks


def heatmap(data):
    # Prepare the data for the heatmap
    metrics = []
    dominant_freqs = []
    confs = []

    if data:
        # nbins = round(data[0]['freq']*(data[0]['t_end'] - data[0]['t_start']))
        nbins = round((data[0]["t_end"] - data[0]["t_start"]))
    for d in data:
        metric = d["metric"]
        if len(d["dominant_freq"]) > 0 and len(d["conf"]) > 0:
            max_conf_index = np.argmax(d["conf"])
            dominant_freq = d["dominant_freq"][max_conf_index]
            conf = d["conf"][max_conf_index] * 100
        else:
            continue
            dominant_freq = 0
            conf = 0
        metrics.append(metric)
        dominant_freqs.append(dominant_freq)
        confs.append(conf)

    # Convert to DataFrame
    heatmap_data = pd.DataFrame(
        {"Metric": metrics, "Dominant Frequency": dominant_freqs, "Confidence": confs}
    )

    # Ensure data is not empty and contains valid ranges
    if heatmap_data.empty:
        raise ValueError("The heatmap data is empty. Cannot generate heatmap.")

    # Define equally spaced bins for Dominant Frequency
    freq_min, freq_max = (
        heatmap_data["Dominant Frequency"].min(),
        heatmap_data["Dominant Frequency"].max(),
    )
    if freq_min == freq_max:
        raise ValueError(
            "The range for Dominant Frequency binning is invalid. Minimum and maximum values are the same."
        )
    freq_bins = np.linspace(freq_min, freq_max, num=nbins)  # Example: 4 bins

    # Bin the Dominant Frequency data
    heatmap_data["Dominant Frequency Binned"] = pd.cut(
        heatmap_data["Dominant Frequency"], bins=freq_bins, include_lowest=True
    )

    # Convert Interval bins to strings for Plotly
    heatmap_data["Dominant Frequency Binned"] = heatmap_data[
        "Dominant Frequency Binned"
    ].astype(str)
    heatmap_data.sort_values(by="Dominant Frequency", inplace=True, ignore_index=True)

    # Pivot the DataFrame to switch x and y axes
    heatmap_pivot = heatmap_data.pivot_table(
        index="Dominant Frequency Binned",
        columns="Metric",
        values="Confidence",
        aggfunc="mean",
        sort=False,
    )
    # or without sorting
    # heatmap_pivot = heatmap_data.pivot_table(index='Dominant Frequency Binned', columns='Metric', values='Confidence', aggfunc='mean')

    # Create the heatmap with switched axes
    fig = px.imshow(
        heatmap_pivot,
        labels=dict(x="Metric", y="Dominant Frequency", color="Confidence"),
        # text_auto=True,
        origin="lower",
        color_continuous_scale="Viridis",
        aspect="auto",  #'auto' or equal
    )
    fig.update_layout(
        # height=1500,
        xaxis_tickangle=-45,
        margin=dict(l=100, r=100, t=50, b=150),
        coloraxis_colorbar=dict(
            yanchor="top",
            y=1,
            ticks="outside",
            ticksuffix=" %",
            # lenmode="pixels",
            # len=200,
        ),
    )
    fig = format_plot_and_ticks(fig,False, True,False,False)
    fig.show()


def scatter(df, x, y, color, symbol) -> None:
    # Create the scatter plot
    fig = px.scatter(
        df, x=x, y=y, color=color, symbol=symbol, color_continuous_scale="Viridis"
    )
    # Display the plot
    fig.update_layout(
        xaxis_title=x,
        yaxis_title=y,
        xaxis_tickangle=-45,
        margin=dict(l=100, r=100, t=50, b=150),
        coloraxis_colorbar=dict(
            orientation="h", ticks="outside", ticksuffix=" %", title=""
        ),
    )
    fig.show()


def scatter2D(df) -> None:
    # Create the scatter plot
    fig = px.scatter(
        df,
        x="Metric",
        y="Dominant Frequency",
        color="Confidence",
        labels={
            "Metric": "Metric",
            "Dominant Frequency": "Dominant Frequency",
            "Confidence": "Confidence",
        },
        color_continuous_scale="Viridis",
    )
    # Display the plot
    fig.update_layout(
        xaxis_title="Metric",
        yaxis_title="Dominant Frequency",
        xaxis_tickangle=-45,
        margin=dict(l=100, r=100, t=50, b=150),
        coloraxis_colorbar=dict(yanchor="top", y=1, ticks="outside", ticksuffix=" %"),
    )
    fig = format_plot_and_ticks(fig,False, True,False,False)
    fig.show()


def heatmap_2(data):
    # Prepare the data for the heatmap
    zscore = True
    metrics = []
    dominant_freqs = []
    confs = []
    counter = 0

    for d in data:
        metric = d["metric"]
        if len(d["dominant_freq"]) > 0 and len(d["conf"]) > 0:
            max_conf_index = np.argmax(d["conf"])
            dominant_freq = d["dominant_freq"][max_conf_index]
            conf = d["conf"][max_conf_index]
        else:
            continue
            dominant_freq = 0
            conf = 0
        metrics.append(metric)
        # metrics.append(f"{counter}")
        dominant_freqs.append(dominant_freq)
        confs.append(conf)
        counter += 1

    # Convert to DataFrame
    heatmap_data = pd.DataFrame(
        {"Metric": metrics, "Dominant Frequency": dominant_freqs, "Confidence": confs}
    )

    # Ensure data is not empty and contains valid ranges
    if heatmap_data.empty:
        raise ValueError("The heatmap data is empty. Cannot generate heatmap.")

    # Select the dominant frequency with the highest confidence for each Metric
    dominant_freq_per_metric = heatmap_data.loc[
        heatmap_data.groupby("Metric")["Confidence"].idxmax()
    ]
    dominant_freq_per_metric.sort_values(
        "Dominant Frequency", ignore_index=True, inplace=True
    )

    # Create a DataFrame for the differences in Dominant Frequency
    # metrics_unique = sorted(dominant_freq_per_metric['Metric'].unique())
    metrics_unique = dominant_freq_per_metric["Metric"].tolist()
    heatmap_diff = pd.DataFrame(index=metrics_unique, columns=metrics_unique)

    if zscore:
        mu = np.mean(dominant_freqs)
        sigma = np.std(dominant_freqs)

    for i, metric_i in enumerate(metrics_unique):
        for j, metric_j in enumerate(metrics_unique):
            if i <= j:  # Only calculate for one half of the matrix
                freq_i = dominant_freq_per_metric.loc[
                    dominant_freq_per_metric["Metric"] == metric_i, "Dominant Frequency"
                ].values
                freq_j = dominant_freq_per_metric.loc[
                    dominant_freq_per_metric["Metric"] == metric_j, "Dominant Frequency"
                ].values
                if len(freq_i) > 0 and len(freq_j) > 0:
                    if zscore:
                        Z_i = (freq_i[0] - mu) / sigma
                        Z_j = (freq_j[0] - mu) / sigma
                        diff = abs(Z_i - Z_j) * 100
                    else:
                        diff = (
                            abs(freq_i[0] - freq_j[0])
                            / ((freq_i[0] + freq_j[0]) / 2)
                            * 100
                        )
                        # diff = abs(freq_i[0] - freq_j[0]) * 100
                else:
                    diff = -1
                heatmap_diff.at[metric_i, metric_j] = diff
                heatmap_diff.at[metric_j, metric_i] = diff  # Mirror the value

    # # Fill the diagonal with zeros (or any other value if necessary)
    # for metric in metrics_unique:
    #     heatmap_diff.at[metric, metric] = 0

    # Convert to numeric type
    # heatmap_diff = heatmap_diff.astype(float)
    # top = float(max(heatmap_diff.max(),100))

    # Ensure the index and columns are correctly set to metrics_unique
    heatmap_diff.index = metrics_unique
    heatmap_diff.columns = metrics_unique
    plot_heatmap(heatmap_diff)

    heatmap_diff = heatmap_diff.fillna(0)  # Replace NaN with 0 or another strategy
    # Apply hierarchical clustering using fastcluster
    linkage_matrix = fastcluster.linkage(
        heatmap_diff, method="average", metric="euclidean"
    )

    sns.clustermap(
        heatmap_diff,
        row_linkage=linkage_matrix,
        col_linkage=linkage_matrix,
        cmap="viridis",
        annot=False,
        xticklabels=metrics_unique,
        yticklabels=metrics_unique,
        figsize=(12, 10),
    )
    plt.show()


def density_heatmap(data) -> None:
    # Prepare the data for the plot
    data_points = []

    # Calculate number of bins based on data range
    if data:
        t_start = data[0].get("t_start", 0)
        t_end = data[0].get("t_end", 100)  # Use a default end value if not provided
        nbins = round((t_end - t_start) / 10)
    else:
        nbins = 30  # Default value if no data is provided

    for d in data:
        if len(d["dominant_freq"]) > 0 and len(d["conf"]) > 0:
            max_conf_index = np.argmax(d["conf"])
            dominant_freq = d["dominant_freq"][max_conf_index]
            conf = d["conf"][max_conf_index] * 100
            data_points.append((d["metric"], dominant_freq, conf))
        else:
            continue

    # Create a DataFrame for the plot
    df = pd.DataFrame(
        data_points, columns=["Metric", "Dominant Frequency", "Confidence"]
    )

    # Create the density heatmap
    fig = px.density_heatmap(
        df,
        x="Metric",
        y="Dominant Frequency",
        z="Confidence",
        labels={
            "Metric": "Metric",
            "Dominant Frequency": "Dominant Frequency",
            "Confidence": "Confidence",
        },
        # color_continuous_scale='Viridis',
        nbinsx=nbins,  # Number of bins along the x-axis
        nbinsy=nbins,  # Number of bins along the y-axis
        marginal_y="histogram",
    )

    # Display the plot
    fig.update_layout(
        xaxis_title="Metric",
        yaxis_title="Dominant Frequency",
        xaxis_tickangle=-45,
        coloraxis_colorbar=dict(
            yanchor="top", y=1, ticks="outside", ticksuffix=" %", title="Confidence (%)"
        ),
        margin=dict(l=100, r=100, t=50, b=150),
    )
    fig.show()


def plot_heatmap(heatmap_diff):

    # Create the heatmap with switched axes
    fig = px.imshow(
        heatmap_diff,
        labels=dict(x="", y="", color="Difference in Dominant Frequency"),
        # text_auto=True,
        origin="lower",
        # width=1200,  # Adjust width as needed
        # height=1000   # Adjust height as needed
        aspect="auto",
        # color_continuous_scale='RdBu'
        color_continuous_scale=[
            (0, "black"),
            (0.1, "red"),
            (0.2, "orange"),
            (0.4, "yellow"),
            (0.6, "white"),
            (0.8, "blue"),
            (1, "purple"),
        ],
    )

    # Update layout to adjust margins and spacing
    fig.update_layout(
        xaxis_title="Metric",
        yaxis_title="Metric",
        coloraxis_colorbar=dict(
            title="Relative deviation (%)",
            yanchor="top",
            y=1,
            ticks="outside",
            ticksuffix=" %",
            # tickvals=[0, critical/2, critical, 2*critical, 1/critical],
        ),
        xaxis=dict(
            tickmode="linear",  # Ensure tick labels are spaced out
            tickangle=90,  # Rotate tick labels if they overlap
        ),
        margin=dict(l=100, r=100, t=50, b=150),  # Adjust margins to give more space
    )
    fig = format_plot_and_ticks(fig,False, True,False,False)
    fig.show()


def plot_timeseries_metrics(metrics, width=None, height=None):
    fig = go.Figure()
    for metric, arrays in metrics.items():
        if len(arrays[0]) > 1:
            fig.add_trace(
                go.Scatter(x=arrays[1], y=arrays[0], mode="lines+markers", name=metric)
            )

    fig.update_layout(
        xaxis_title="Time (s)",
        yaxis_title="Metrics",
    )
    if width and height:
        fig.update_layout(autosize=False, width=width, height=height)
    fig.show()
    return fig



import sys
import signal
from rich.console import Console
from ftio.api.gekkoFs.jit.jitsettings import JitSettings
from ftio.api.gekkoFs.jit.jittime import JitTime
from ftio.api.gekkoFs.jit.setup_core import (
    post_call,
    pre_call,
    stage_in,
    stage_out,
    start_application,
    start_cargo,
    start_ftio,
    start_gekko_daemon,
    start_gekko_proxy,
)
from ftio.api.gekkoFs.jit.setup_helper import (
    get_address_cargo,
    get_address_ftio,
    handle_sigint,
    hard_kill,
    parse_options,
    cancel_jit_jobs,
    allocate,
    # set_env,
    log_dir,
    print_settings,
    save_bandwidth,
    set_dir_gekko,
    soft_kill,
)


console = Console()


def main() -> None:
    """Executes the jit script
    """

    console.print("\n\n[bold green]################ JIT ################[/]\n")
    settings = JitSettings()
    runtime = JitTime()

    # Parse options
    parse_options(settings, sys.argv[1:])

    # Handle kill signal
    signal.signal(signal.SIGINT, lambda signal, frame: handle_sigint(settings))

    # Clean other jobs
    cancel_jit_jobs(settings)

    # # 1.0 set env variables
    # set_env(settings)

    # 1.1 Allocate resources
    allocate(settings)

    # 1.2 Create folder for logs
    log_dir(settings)
    
    # 1.3 Get the address
    get_address_ftio(settings)

    # 1.4 Get address carg
    get_address_cargo(settings)

    # 1.5 Set Gekko Root dir
    set_dir_gekko(settings)

    # 1.6 Print settings
    print_settings(settings)

    # 2.0 Start Gekko Server (Daemon)
    start_gekko_daemon(settings)

    # 2.1 Start Proxy
    start_gekko_proxy(settings)

    # 3.0 Start Cargo Server
    start_cargo(settings)

    # 4.0 Stage in
    stage_in(settings, runtime)

    # 5.0 Start FTIO
    start_ftio(settings)

    # 6.0 Pre- and application with Gekko intercept
    pre_call(settings)
    start_application(settings, runtime)

    # 7.0 Stage out
    stage_out(settings, runtime)

    # 8.0 Post call if exists
    post_call(settings)

    # 9.0 Display total time
    # _ = runtime.print_time()
    runtime.print_and_save_time(settings)

    # save_bandwidth
    save_bandwidth(settings)

    # 11.0 Soft kill
    soft_kill(settings)

    # 12.0 Hard kill
    hard_kill(settings)

    console.print("[bold green]############### JIT completed ###############[/]")
    sys.exit(0)


if __name__ == "__main__":
    main()
